###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""An agent to extend MC productions based on the remaining events to
produce."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import math
import datetime

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Agent.MCExtensionAgent import MCExtensionAgent as DIRACMCExtensionAgent
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import getCPUNormalizationFactorAvg, \
    getEventsToProduce, \
    getProductionParameterValue

AGENT_NAME = 'Transformation/MCExtensionAgent'


class MCExtensionAgent(DIRACMCExtensionAgent):
  """MCExtensionAgent."""

  def __init__(self, *args, **kwargs):
    """c'tor."""
    DIRACMCExtensionAgent.__init__(self, *args, **kwargs)

    self.rpcProductionRequest = None
    self.transClient = None

    self.enableFlag = True

    # default values
    self.cpuE = 1
    self.cpuTimeAvg = 200000
    self.cpuNormalizationFactorAvg = 1.0

    # Artificial boost of the number of events requested to be created
    self.extensionFactorBoost = 20  # Meaning 20% more than what is calculated

  #############################################################################
  def initialize(self):
    """Logs some parameters and initializes the clients."""
    self.extensionFactorBoost = self.am_getOption('extensionFactorBoost', self.extensionFactorBoost)

    self.rpcProductionRequest = ProductionRequestClient()
    self.transClient = TransformationClient()

    self.log.info('Will consider the following transformation types: %s' % str(self.transformationTypes))
    self.log.info('Will create a maximum of %s tasks per iteration' % self.maxIterationTasks)

    return S_OK()

  #############################################################################
  def execute(self):
    """The MCExtensionAgent execution method."""

    self.enableFlag = self.am_getOption('EnableFlag', 'True')
    if not self.enableFlag == 'True':
      self.log.info("MCExtensionAgent is disabled by configuration option EnableFlag")
      return S_OK('Disabled via CS flag')

    # done every cycle, as they may have changed
    self._getCPUParameters()

    # get the production requests in which we are interested
    productionRequests = self.rpcProductionRequest.getProductionRequestSummary('Active', 'Simulation')
    if productionRequests['OK']:
      productionRequests = productionRequests['Value']
      self.log.info("Requests considered: %s" % ', '.join([str(prod) for prod in productionRequests]))
    else:
      message = "RPC call to ProductionRequest service failed : %s" % productionRequests['Message']
      self.log.error(message)
      return S_ERROR(message)

    for productionRequestID, productionRequestSummary in productionRequests.items():
      ret = self._checkProductionRequest(productionRequestID, productionRequestSummary)
      if not ret['OK']:
        return ret

    return S_OK()

  #############################################################################

  def _getCPUParameters(self):
    """Get the CPUTimeAvg and CPUNormalizationFactorAvg from config, or as a
    fail-over, there are some defaults."""

    op = Operations()
    self.cpuTimeAvg = op.getValue('Transformations/cpuTimeAvg', self.cpuTimeAvg)
    self.log.verbose("cpuTimeAvg = %d" % self.cpuTimeAvg)

    try:
      self.cpuNormalizationFactorAvg = getCPUNormalizationFactorAvg()
      self.log.verbose("cpuNormalizationFactorAvg = %d" % self.cpuNormalizationFactorAvg)
    except RuntimeError:
      self.log.info(
          "Could not get CPUNormalizationFactorAvg from config, defaulting to %d" %
          self.cpuNormalizationFactorAvg)

  #############################################################################
  def _checkProductionRequest(self, productionRequestID, productionRequestSummary):
    """Check if a production request need to be extended and do it if
    needed."""

    # check if enough events have been produced
    missingEvents = productionRequestSummary['reqTotal'] - productionRequestSummary['bkTotal']
    self.log.info("Missing events for production request %d: %d" % (productionRequestID, missingEvents))
    if productionRequestSummary['bkTotal'] > 0 and missingEvents <= 0:
      message = "Enough events produced for production request %d" % productionRequestID
      self.log.verbose(message)
      return S_OK(message)

    # get the associated productions/transformations progress
    productionsProgress = self.rpcProductionRequest.getProductionProgressList(int(productionRequestID))
    if productionsProgress['OK']:
      productionsProgress = productionsProgress['Value']
    else:
      message = 'Failed to get productions progress : %s' % productionsProgress['Message']
      self.log.error(message)
      return S_ERROR(message)
    productionsProgress = productionsProgress['Rows']
    self.log.verbose("Progress for production request %d: %s" % (productionRequestID, str(productionsProgress)))

    # get the informations for the productions/transformations
    productions = []
    simulation = None
    for productionProgress in productionsProgress:
      productionID = productionProgress['ProductionID']
      production = self.transClient.getTransformation(productionID)
      if not production['OK']:
        message = 'Failed to get informations on production %d : %s' % (productionID, production['Message'])
        self.log.error(message)
        return S_ERROR(message)
      production = production['Value']
      productions.append(production)

      # determine which one is the simulation production
      if production['Type'] in self.transformationTypes:
        simulation = production
        simulationID = productionID
        for prodProgress in productionsProgress:
          if prodProgress['ProductionID'] == simulationID:
            simulationProgress = prodProgress
            self.log.info("Progress for the simulation production %d of request %d: %s" % (simulationID,
                                                                                           productionRequestID,
                                                                                           str(simulationProgress)))

    if simulation is None:
      message = 'Failed to get simulation production for request %d' % productionRequestID
      self.log.error(message)
      return S_ERROR(message)

    if simulation['Status'].lower() != 'idle':
      # the simulation is still producing events
      message = "Simulation for production request %d is not Idle (%s)" % (productionRequestID, simulation['Status'])
      self.log.verbose(message)
      return S_OK(message)

    # Checking how long ago this production became 'Idle'
    res = self.transClient.getTransformationLogging(simulationID)
    if not res['OK']:
      return res
    lastLoggingEntry = res['Value'][-1]
    if ('idle' in lastLoggingEntry['Message'].lower()) and \
       ((datetime.datetime.utcnow() - lastLoggingEntry['MessageDate']).seconds < 900):
      self.log.verbose("Prod %d is in 'Idle' for less than 15 minutes, waiting a bit" % simulationID)
      return S_OK("Prod %d is in 'Idle' for less than 15 minutes, waiting a bit" % simulationID)

    if simulationProgress['BkEvents'] < productionRequestSummary['reqTotal']:
      # the number of events produced by the simulation is of the order of the number of events requested
      # -> there is probably no stripping production, no extension factor necessary
      return self._extendProduction(simulation, 1.0, missingEvents)
    else:
      # the number of events produced by the simulation is more than the number
      # of events requested, yet events are missing -> there is probably a
      # stripping production, an extension factor is needed to account for
      # stripped events some events may still be processed (eg. merged), so
      # wait that all the productions are idle
      if all(production['Status'].lower() == 'idle' for production in productions):
        try:
          extensionFactor = float(simulationProgress['BkEvents']) / float(productionRequestSummary['bkTotal'])
          return self._extendProduction(simulation, extensionFactor, missingEvents)
        except ZeroDivisionError:
          return S_OK()
      else:
        return S_OK()

  #############################################################################
  def _extendProduction(self, production, extensionFactor, eventsNeeded):
    """Extends a production to produce eventsNeeded*extensionFactor more
    events."""
    productionID = production['TransformationID']

    cpuEProd = getProductionParameterValue(production['Body'], 'CPUe')
    if cpuEProd is None:
      self.log.warn("CPUe for transformation %d is not set, skipping for now" % productionID)
      return S_OK()
    cpuE = int(round(float(cpuEProd)))

    self.log.info("Extending production %d, that is still missing %d events. \
                    Extension factor = %d, boost = %d" % (productionID,
                                                          eventsNeeded,
                                                          extensionFactor,
                                                          self.extensionFactorBoost))

    eventsToProduce = eventsNeeded * extensionFactor * (float(100 + self.extensionFactorBoost) / 100)
    max_e = getEventsToProduce(cpuE, self.cpuTimeAvg, self.cpuNormalizationFactorAvg)
    numberOfTasks = int(math.ceil(float(eventsToProduce) / float(max_e)))
    self.log.info("Extending production %d by %d tasks" % (productionID, numberOfTasks))

    # extend the transformation by the determined number of tasks
    res = self.transClient.extendTransformation(productionID, numberOfTasks)
    if not res['OK']:
      message = 'Failed to extend transformation %d : %s' % (productionID, res['Message'])
      self.log.error(message)
      return S_ERROR(message)
    else:
      message = "Successfully extended transformation %d by %d tasks" % (productionID, numberOfTasks)
      self.log.info(message)

      res = self.transClient.setTransformationParameter(productionID, 'Status', 'Active')
      if not res['OK']:
        message = 'Failed to set transformation %d to Active' % productionID
        self.log.error(message)
        return S_ERROR(message)

      return S_OK(message)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Show replicas for a (set of) LFNs."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerFileSwitches()
  Script.registerSwitch('a', "All", "  Also show inactive replicas")
  Script.registerSwitch('', 'DiskOnly', '  Show only disk replicas')
  Script.registerSwitch('', 'PreferDisk', "  If disk replica, don't show tape replicas")
  Script.registerSwitch('', 'ForJobs', '  Select only replicas that can be used for jobs')

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] [<LFN>] [<LFN>...]' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=False)

  from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeLfnReplicas
  from DIRAC import exit
  exit(executeLfnReplicas(dmScript))


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript
  from DIRAC.Core.Base import Script

  dmScript = DMScript()
  dmScript.registerBKSwitches()
  dmScript.registerFileSwitches()

  fixIt = False
  Script.registerSwitch('', 'FixIt', 'Set replicas problematic if needed')
  Script.setUsageMessage("""
  Check the integrity of the state of the storages and information in the File Catalogs
  for a given file or a collection of files.

  Usage:
     %s <lfn | fileContainingLfns> <SE> <status>
  """ % Script.scriptName)

  Script.parseCommandLine()

  for opt, val in Script.getUnprocessedSwitches():
    if opt == 'FixIt':
      fixIt = True

  from DIRAC import gLogger
  gLogger.setLevel('INFO')
  from LHCbDIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient

  for lfn in Script.getPositionalArgs():
    dmScript.setLFNsFromFile(lfn)
  lfns = dmScript.getOption('LFNs')
  if not lfns:
    print("No LFNs given...")
    Script.showHelp(exitCode=1)

  integrityClient = DataIntegrityClient()
  res = integrityClient.catalogFileToBK(lfns)
  if not res['OK']:
    gLogger.error(res['Message'])
    DIRAC.exit(1)
  replicas = res['Value']['CatalogReplicas']
  metadata = res['Value']['CatalogMetadata']
  res = integrityClient.checkPhysicalFiles(replicas, metadata, fixIt=fixIt)
  if not res['OK']:
    gLogger.error(res['Message'])
    DIRAC.exit(1)
  DIRAC.exit(0)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Minimal test for TransformationCleaningAgent."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib
import unittest

from mock import MagicMock
from DIRAC import S_OK, gLogger

from LHCbDIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent


class TransformationCleaningAgentTestCase(unittest.TestCase):

  def setUp(self):
    self.mockAM = MagicMock()
    self.mockAM.am_getOption.return_value = S_OK()
    self.agent = importlib.import_module('LHCbDIRAC.TransformationSystem.Agent.TransformationCleaningAgent')
    self.agent.AgentModule = self.mockAM
    self.mockDTCAObj = MagicMock()
    self.mockDTCAObj.getTransformationDirectories.return_value = {
        'OK': True, 'Value': ['/this/is/SIM/dir', '/this/is/FOO/dir', '/this/is/BAR/dir']}
    self.mockDTCA = MagicMock()
    self.mockDTCA.return_value = self.mockDTCAObj
    self.agent.DiracTCAgent = self.mockDTCA
    self.agent = TransformationCleaningAgent('a', 'b')
    self.agent.log = gLogger
    self.mockSUC = MagicMock()
    self.mockSUC.getStorageDirectories.return_value = {'OK': True, 'Value': []}
    self.agent.storageUsageClient = self.mockSUC

  def test_getTransformationDirectories(self):
    res = self.agent.getTransformationDirectories(1)
    self.assertTrue(res['OK'])


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TransformationCleaningAgentTestCase)
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DIRAC ProductionRequestDB class is a front-end to the repository database
containing Production Requests and other related tables."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# Defined states:
# 'New'
# 'BK OK'
# 'Rejected'
# 'BK Check'
# 'Submitted'
# 'PPG OK'
# 'On-hold'
# 'Tech OK'
# 'Accepted'
# 'Active'
# 'Completed'
# 'Done'
# 'Cancelled'

import json
import time
import threading

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB

from LHCbDIRAC.ProductionManagementSystem.Utilities.Utils import informPeople

__RCSID__ = "$Id$"


class ProductionRequestDB(DB):
  """DB class for ProductionManagement/ProductionRequestDB."""

  def __init__(self):
    """Constructor."""
    DB.__init__(self, 'ProductionRequestDB', 'ProductionManagement/ProductionRequestDB')
    self.dateColumns = ['StartingDate', 'FinalizationDate']
    self.dateFormat = '%Y-%m-%d'
    self.lock = threading.Lock()

# Production Requests table
  requestFields = ['RequestID', 'ParentID', 'MasterID', 'RequestAuthor',
                   'RequestName', 'RequestType', 'RequestState',
                   'RequestPriority', 'RequestPDG', 'RequestWG',
                   'SimCondition', 'SimCondID', 'SimCondDetail',
                   'ProPath', 'ProID', 'ProDetail',
                   'EventType', 'NumberOfEvents', 'Description', 'Comments',
                   'Inform', 'RealNumberOfEvents', 'IsModel', 'Extra',
                   'RetentionRate', 'FastSimulationType', 'StartingDate', 'FinalizationDate',
                   'HasSubrequest', 'bk', 'bkSrTotal', 'bkTotal',  # runtime
                   'rqTotal', 'crTime', 'upTime']  # runtime

  historyFields = ['RequestID', 'RequestState', 'RequestUser', 'TimeStamp']

  # !!! current _escapeValues is buggy !!! None and not using connection...
  # _insert use it, so I can't...
  def _fixedEscapeValues(self, inValues, escape=None):
    """This method used to insert null value to the db, if the inserted value
    is None.

    :param list inValues: list of values
    :param int escape: the index of the value, which will be not escaped.
    """
    result = self._escapeValues(inValues)
    if not result['OK']:
      return result
    outValues = result['Value']
    for i, x in enumerate(outValues):
      if x == 'None' or str(x) == '' or x == '"None"':
        if escape and i == escape:
          continue
        outValues[i] = 'NULL'
    return S_OK(outValues)

  @staticmethod
  def __prefixComments(update, old, user):
    """Add Log style prefix to the record like change."""
    if not update:
      return update
    if not old:
      old = ''
    if not update.startswith(old) and not update.endswith(old):
      return update
    prefix = "Comment by %s on %s: " % (user, time.strftime("%b %d, %Y"))
    if update.startswith(old):
      if old.rstrip():
        old = old.rstrip() + '\n\n'
      return old + prefix + update[len(old):].lstrip()
    return prefix + update.lstrip()

  def __getRequestInfo(self, iD, connection):
    """Retrive info fields from specified ID Used to get ParentID information.

    id must be checked before
    NOTE: it does self.lock.release() in case of errors
    """
    inFields = ['RequestState', 'ParentID', 'MasterID', 'RequestAuthor', 'Inform', 'IsModel']
    result = self._query("SELECT %s " % ','.join(inFields) +
                         "FROM ProductionRequests " +
                         "WHERE RequestID=%s;" % iD, connection)
    if not result['OK']:
      self.lock.release()
      return result
    if len(result['Value']) == 0:
      self.lock.release()
      return S_ERROR('Request does not exist')
    return S_OK(dict(zip(inFields, result['Value'][0])))

  def __getStateAndAuthor(self, iD, connection):
    ''' Return state, Author and inform list of Master for id (or id's own if no parents)
        id must be checked before
        NOTE: it does self.lock.release() in case of errors
    '''
    result = self.__getRequestInfo(iD, connection)
    if not result['OK']:
      return result
    pinfo = result['Value']
    if not pinfo['MasterID']:
      return S_OK([pinfo['RequestState'], pinfo['RequestAuthor'], pinfo['Inform'], pinfo['IsModel']])
    result = self.__getRequestInfo(pinfo['MasterID'], connection)
    if not result['OK']:
      return result
    pinfo = result['Value']
    return S_OK([pinfo['RequestState'], pinfo['RequestAuthor'], pinfo['Inform'], pinfo['IsModel']])

  def __checkMaster(self, master, iD, connection):
    """Return State of Master for id (or id's own if no parents) id and master
    must be checked before.

    It check that master can
    be reached with ParentID links.
    NOTE: it does self.lock.release() in case of errors
    """
    while True:
      result = self.__getRequestInfo(iD, connection)
      if not result['OK']:
        return result
      pinfo = result['Value']
      if iD == master:
        return S_OK(pinfo['RequestState'])
      if pinfo['MasterID'] != master:
        self.lock.release()
        return S_ERROR('Wrong MasterID for this ParentID')
      if not pinfo['ParentID'] or pinfo['ParentID'] == iD:
        self.lock.release()
        return S_ERROR('Parent tree is broken. Please contact expert')
      iD = pinfo['ParentID']

  def createProductionRequest(self, requestDict, creds):
    ''' Create new Production Request
        TODO: Complete check of content
    '''

    rec = dict.fromkeys(self.requestFields[1:-9], None)

    for x in requestDict:
      if x in rec and str(requestDict[x]) != '':
        rec[x] = requestDict[x]  # set only known not empty fields
    if rec['NumberOfEvents']:  # Set RealNumberOfEvents if specified
      try:
        num = int(rec['NumberOfEvents'])
        if num > 0:
          rec['RealNumberOfEvents'] = num
      except ValueError:
        pass
      except TypeError:
        gLogger.warn("RealNumberOfEvents not defined")
    if not rec['MasterID']:
      rec['RequestPDG'] = ''
      if not rec['RequestState']:
        rec['RequestState'] = 'New'
    else:
      rec['RequestPDG'] = None
      rec['RequestState'] = None

    if rec['RequestState']:
      if not rec['RequestState'] in ['New', 'BK Check', 'Submitted']:
        return S_ERROR("The request can't be created in '%s' state" % rec['requestState'])
      if rec['RequestState'] != 'New':
        # !!! full information check must be here, but currently in the JS...
        # so we only check EventType consistency
        if not rec['EventType']:
          return S_ERROR("Please specify Event type/number or add subrequest(s)")
    if 'Comments' in rec:
      rec['Comments'] = self.__prefixComments(rec['Comments'], '', creds['User'])
    rec['IsModel'] = 0

    recl = [rec[x] for x in self.requestFields[1:-9]]
    result = self._fixedEscapeValues(recl)
    if not result['OK']:
      return result
    recls = result['Value']
    if recl[-1] == "NULL":  # This is FastSimulationType, which can be None
      recls[-1] = '"None"'

    for dateValues in self.dateColumns:
      recls.append("STR_TO_DATE('%s','%s')" %
                   (requestDict.get(dateValues, time.strftime(self.dateFormat)), self.dateFormat))

    self.lock.acquire()  # transaction begin ?? may be after connection ??
    result = self._getConnection()
    if not result['OK']:
      self.lock.release()
      return S_ERROR('Failed to get connection to MySQL: ' + result['Message'])
    connection = result['Value']

    if rec['MasterID']:  # have to check ParentID and MasterID consistency
      try:
        masterID = int(rec['MasterID'])
      except ValueError:
        self.lock.release()
        return S_ERROR('MasterID is not a number')
      if not rec['ParentID']:
        self.lock.release()
        return S_ERROR('MasterID can not be without ParentID')
      try:
        parentID = int(rec['ParentID'])
      except ValueError:
        self.lock.release()
        return S_ERROR('ParentID is not a number')
      result = self.__checkMaster(masterID, parentID, connection)
      if not result['OK']:
        return result
      result = self.__getStateAndAuthor(masterID, connection)
      if not result['OK']:
        return result
      requestState, requestAuthor, _requestInform, isModel = result['Value']
      if requestState != 'New':
        self.lock.release()
        return S_ERROR("Requests can't be modified after submission")
      if requestAuthor != creds['User'] and not (isModel and creds['Group'] == 'lhcb_tech'):
        self.lock.release()
        return S_ERROR("Only request author can add subrequests")
    elif rec['ParentID']:
      try:
        parentID = int(rec['ParentID'])
      except ValueError:
        self.lock.release()
        return S_ERROR('ParentID is not a number')
      result = self.__getStateAndAuthor(parentID, connection)
      if not result['OK']:
        return result
    result = self.__checkIOTypes(requestDict)
    if not result['OK']:
      self.lock.release()
      return result

    req = "INSERT INTO ProductionRequests ( " + ','.join(self.requestFields[1:-7])
    req += " ) VALUES ( %s );" % ','.join(recls)
    result = self._update(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    req = "SELECT LAST_INSERT_ID();"
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    requestID = int(result['Value'][0][0])

    # Update history for masters. Errors are not reported back to the user.
    if rec['RequestState']:
      result = self._update("INSERT INTO RequestHistory (" +
                            ','.join(self.historyFields[:-1]) +
                            ") VALUES ( %s,'%s','%s')" %
                            (requestID, str(rec['RequestState']), str(rec['RequestAuthor'])), connection)
      if not result['OK']:
        gLogger.error(result['Message'])
    self.lock.release()
    if rec['RequestState'] in ['BK Check', 'Submitted']:
      rec['RequestID'] = requestID
      informPeople(rec, '', rec['RequestState'], creds['User'], rec['Inform'])
    return S_OK(requestID)

  def __checkIOTypes(self, requestDict):
    """Check the input type of each step matches the output of a previous
    step."""
    prodDetail = requestDict.get('ProDetail')
    if prodDetail is not None:
      try:
        proDetail = json.loads(prodDetail)
      except Exception:
        return S_ERROR('Content of ProDetail field cannot be loaded')
      for i in range(20):
        outputKey = 'p' + str(i) + 'OFT'
        inputKey = 'p' + str(i + 1) + 'IFT'
        if outputKey in proDetail and inputKey in proDetail:
          inputFileTypes = set(proDetail[inputKey].replace('(N)', '').replace('(Y)', '').split(','))
          outputFileTypes = set(proDetail[outputKey].replace('(N)', '').replace('(Y)', '').split(','))
          if not inputFileTypes.issubset(outputFileTypes):
            return S_ERROR('Input for step ' + str(i + 1) + ' does not match the output of step ' + str(i))
    return S_OK()

  @staticmethod
  def __addMonitoring(req, order):
    """Append monitoring columns.

    Somehow tricky SQL. Most probable need optimizations, but ok for
    now.
    """
    rQuery = "SELECT t.*,MIN(rh.TimeStamp) AS crTime,"
    rQuery += "           MAX(rh.TimeStamp) AS upTime "
    rQuery += "FROM "
    rQuery += "(SELECT t.*,CAST(COALESCE(SUM(sr.RealNumberOfEvents),0)+"
    rQuery += "                COALESCE(t.RealNumberOfEvents,0) AS SIGNED)"
    rQuery += "           AS rqTotal "
    rQuery += " FROM "
    rQuery += " (SELECT t.*,CAST(COALESCE(SUM(t.bkSrTotal),0)+"
    rQuery += "                  COALESCE(t.bk,0) AS SIGNED) AS bkTotal FROM "
    rQuery += "  (SELECT t.*,CAST(LEAST(COALESCE(SUM(pp.BkEvents),0),"
    rQuery += "                   COALESCE(SUM(sr.RealNumberOfEvents),0)) AS SIGNED)"
    rQuery += "              AS bkSrTotal FROM "
    rQuery += "   (SELECT t.*, CAST(SUM(pp.BkEvents) AS SIGNED)"
    rQuery += "                AS bk FROM (%s) as t " % req
    rQuery += "    LEFT JOIN ProductionProgress as pp ON (pp.RequestID=t.RequestID "
    rQuery += "    AND pp.Used=1) GROUP BY t.RequestID) as t "
    rQuery += "   LEFT JOIN ProductionRequests AS sr ON t.RequestID=sr.MasterID "
    rQuery += "   LEFT JOIN ProductionProgress AS pp ON (sr.RequestID=pp.RequestID "
    rQuery += "   AND pp.Used=1) GROUP BY t.RequestID) AS t"
    rQuery += "  GROUP BY t.RequestID) AS t"
    rQuery += " LEFT JOIN ProductionRequests as sr ON sr.MasterID=t.RequestID "
    rQuery += " GROUP BY t.RequestID) as t"
    rQuery += " LEFT JOIN RequestHistory as rh ON rh.RequestID=t.RequestID "
    rQuery += " GROUP BY t.RequestID"

    return rQuery + order

  def getProductionRequest(self, requestIDList, subrequestsFor=0,
                           sortBy='', sortOrder='ASC',
                           offset=0, limit=0, filterIn=None):
    """Get the Production Request(s) details.

    If requestIDList is not empty, only productions from the list are
    returned. Otherwise master requests are returned (without
    subrequests) or all subrequests of 'subrequestsFor' (when
    specified). Parameters with explicit types are assumed checked by
    service.
    """
    if filterIn is None:
      filterIn = {}
    try:  # test parameters
      for x in requestIDList:
        y = int(x)
    except ValueError:
      return S_ERROR("Bad parameters (all request IDs must be numbers)")
    idFilter = False
    try:  # test filters
      sfilter = []
      for x in filterIn:
        if x not in self.requestFields[:-7]:
          return S_ERROR("bad field in filterIn")
        val = str(filterIn[x])
        if val:
          val = "( " + ','.join(['"' + y + '"' for y in val.split(',')]) + ") "
          if x == "RequestID":
            # Collect also masters masters
            req = "SELECT RequestID,MasterID FROM ProductionRequests "
            req += "WHERE RequestID IN %s" % val
            result = self._query(req)
            if not result['OK']:
              return result
            val = []
            for y in result['Value']:
              if str(y[1]) != 'None':
                val.append(str(y[1]))
              else:
                val.append(str(y[0]))
            if not val:
              return S_OK({'Rows': [], 'Total': 0})
            val = "( " + ','.join(val) + ") "
            idFilter = True
          sfilter.append(" t.%s IN %s " % (x, val))
      sfilter = " AND ".join(sfilter)
    except Exception as e:
      return S_ERROR("Bad filter content " + str(e))

    if sortBy:
      if sortBy not in self.requestFields[:-7]:
        return S_ERROR("sortBy field does not exist")
      if sortOrder != 'ASC':
        sortOrder = 'DESC'

    fields = ','.join(['t.' + x for x in self.requestFields[:-7]])
    req = "SELECT %s ,COUNT(sr.RequestID) AS HasSubrequest " % fields
    req += "FROM ProductionRequests as t "
    req += "LEFT JOIN ProductionRequests AS sr ON t.RequestID=sr.ParentID "
    req += "WHERE "

    if requestIDList:
      idlist = ','.join([str(x) for x in requestIDList])
      where = "t.RequestID IN (%s)" % idlist
    else:
      if subrequestsFor:
        where = "t.ParentID=%s" % subrequestsFor
      else:
        where = sfilter
        if not idFilter:
          if where:
            where += " AND t.ParentID IS NULL"
          else:
            where = "t.ParentID IS NULL"
    req += where
    req += " GROUP BY t.RequestID"
    order = ""
    if sortBy:
      # order have to be applyed twice: before LIMIT and at the end
      order = " ORDER BY %s %s" % (sortBy, sortOrder)
      req += order
    if limit and not subrequestsFor:
      req += " LIMIT %s,%s" % (offset, limit)
    result = self._query(self.__addMonitoring(req, order))
    if not result['OK']:
      return result

    rows = [dict(zip(self.requestFields, row)) for row in result['Value']]
    total = len(rows)
    if limit:
      result = self._query("SELECT COUNT(*) FROM ProductionRequests AS t WHERE %s" % where)
      if not result['OK']:
        return result
      total = result['Value'][0][0]
    return S_OK({'Rows': rows, 'Total': total})

  def __checkUpdate(self, update, old, creds, connection):
    """Check that update is possible.

    Return dict with values for _inform_people (with
    state=='' in  case notification is not required)
    NOTE: unlock in case of errors
    """
    requestID = old['RequestID']
    result = self.__getStateAndAuthor(requestID, connection)
    if not result['OK']:
      return result
    requestState, requestAuthor, requestInform, isModel = result['Value']
    rec = old.copy()
    rec.update(update)
    inform = {'rec': rec, 'state': '', 'author': str(requestAuthor),
              'oldstate': requestState, 'inform': requestInform}

    hasSubreq = False
    if not old['MasterID']:
      result = self._query("SELECT RequestID " +
                           "FROM ProductionRequests " +
                           "WHERE MasterID=%s" % requestID, connection)
      if not result['OK']:
        self.lock.release()
        return result
      if result['Value']:
        hasSubreq = True

    if creds['Group'] in ['diracAdmin', 'lhcb_admin']:
      return S_OK(inform)

    # Tech expert can (un)mark any master request as model. But only explicitly.
    if creds['Group'] == 'lhcb_tech' and not old['MasterID']:
      if len(update) == 1 and 'IsModel' in update:
        if str(update['IsModel']) == '1':
          update['IsModel'] = 1
        else:
          update['IsModel'] = 0
        return S_OK(inform)
    else:
      if 'IsModel' in update:
        del update['IsModel']

    if requestState in ['Done', 'Cancelled'] and (creds['Group'] not in ['diracAdmin', 'lhcb_admin']):
      self.lock.release()
      return S_ERROR("Done or cancelled requests can't be modified")

    # Check that a person can update in general (that also means he can
    # change at least comments)
    if requestState in ['New', 'BK OK', 'Rejected']:
      if requestAuthor != creds['User'] and not (isModel and creds['Group'] == 'lhcb_tech'):
        self.lock.release()
        return S_ERROR("Only author is allowed to modify unsubmitted request")
    elif requestState == 'BK Check':
      if creds['Group'] != 'lhcb_bk':
        self.lock.release()
        return S_ERROR("Only BK expert can manage new Simulation Conditions")
    elif requestState == 'Submitted':
      if creds['Group'] != 'lhcb_ppg' and creds['Group'] != 'lhcb_tech':
        self.lock.release()
        return S_ERROR("Only PPG members or Tech. experts are allowed to sign submitted request")
    elif requestState == 'PPG OK':
      if creds['Group'] != 'lhcb_tech':
        self.lock.release()
        return S_ERROR("Only Tech. experts are allowed to sign this request")
    elif requestState == 'On-hold':
      if creds['Group'] != 'lhcb_tech':
        self.lock.release()
        return S_ERROR("Only Tech. experts are allowed to sign this request")
    elif requestState == 'Tech OK':
      if creds['Group'] != 'lhcb_ppg':
        self.lock.release()
        return S_ERROR("Only PPG members are allowed to sign this request")
    elif requestState == 'Accepted':
      if creds['Group'] != 'lhcb_prmgr':
        self.lock.release()
        return S_ERROR("Only Tech. experts are allowed to manage accepted request")
    elif requestState in ['Active', 'Completed']:
      if not creds['Group'] in ['lhcb_prmgr', 'lhcb_prod']:
        self.lock.release()
        return S_ERROR("Only experts are allowed to manage active request")
    elif requestState in ['Done', 'Cancelled']:
      if creds['Group'] not in ['diracAdmin', 'lhcb_admin']:
        self.lock.release()
        return S_ERROR("Only admin can violate the system logic")
    else:
      self.lock.release()
      return S_ERROR("The request is in unknown state '%s'" % requestState)

    if old['MasterID']:  # for subrequests it's simple
      if requestState == 'New':
        for x in update:
          if x not in ['EventType', 'NumberOfEvents', 'RealNumberOfEvents', 'Comments']:
            self.lock.release()
            return S_ERROR("%s is not allowed in subrequests" % x)
          if x != 'Comments' and not update[x]:
            self.lock.release()
            return S_ERROR("You must specify event type and number")
      else:
        if len(update) != 1 or 'Comments' not in update:
          self.lock.release()
          return S_ERROR("Only comments can be changed for subrequest in progress")
      return S_OK(inform)
    # for masters it is more complicated...
    if requestState == 'New':
      if 'RequestState' not in update:
        return S_OK(inform)
      if update['RequestState'] in ['BK Check', 'Submitted']:
        # !!! full information check must be here, but currently in the JS...
        # so we only check EventType consistency
        eventType = old['EventType']
        if 'EventType' in update:
          eventType = update['EventType']
        # gLogger.error(str(update))
        if eventType and hasSubreq:
          self.lock.release()
          return S_ERROR("The request has subrequests, so it must not specify Event type")
        if not eventType and not hasSubreq:
          self.lock.release()
          return S_ERROR("Please specify Event type/number or add subrequest(s)")
      else:
        self.lock.release()
        _msgTuple = (requestState, update['RequestState'])
        return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
    elif requestState == 'Rejected':
      if len(update) != 1 or update.get('RequestState', '') != 'New':
        self.lock.release()
        return S_ERROR("Rejected requests must be resurrected before modifications")
    elif requestState == 'BK Check':
      for x in update:
        if x not in ['RequestState', 'SimCondition', 'SimCondID', 'SimCondDetail', 'Comments', 'Inform', 'Extra']:
          self.lock.release()
          return S_ERROR("%s can't be modified during BK check" % x)
      if 'RequestState' not in update:
        return S_OK(inform)
      if not update['RequestState'] in ['BK OK', 'Rejected']:
        self.lock.release()
        _msgTuple = (requestState, update['RequestState'])
        return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
      if update['RequestState'] == 'BK OK' and not update.get('SimCondID', old['SimCondID']):
        self.lock.release()
        return S_ERROR("Registered simulation conditions required to sign for BK OK")
    elif requestState == 'BK OK':
      for x in update:
        if x not in ['RequestState', 'Comments', 'Inform', 'Extra']:
          self.lock.release()
          return S_ERROR("%s can't be modified after BK check" % x)
      if 'RequestState' not in update:
        return S_OK(inform)
      if not update['RequestState'] in ['Submitted', 'Rejected']:
        self.lock.release()
        _msgTuple = (requestState, update['RequestState'])
        return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
    elif requestState == 'Submitted':
      if creds['Group'] == 'lhcb_ppg':
        for x in update:
          if x not in ('RequestState', 'RequestWG', 'Comments', 'Inform', 'RequestPriority', 'Extra',
                       'StartingDate', 'FinalizationDate', 'RetentionRate'):
            self.lock.release()
            return S_ERROR("%s can't be modified during PPG signing" % x)
        if 'RequestState' not in update:
          return S_OK(inform)
        if update['RequestState'] == 'Accepted':
          update['RequestState'] = 'PPG OK'
        if not update['RequestState'] in ['PPG OK', 'Rejected']:
          self.lock.release()
          _msgTuple = (requestState, update['RequestState'])
          return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
      if creds['Group'] == 'lhcb_tech':
        for x in update:
          if x not in ['RequestState', 'Comments', 'Inform', 'ProPath', 'ProID', 'ProDetail', 'Extra']:
            self.lock.release()
            return S_ERROR("%s can't be modified during Tech signing" % x)
        if 'RequestState' not in update:
          return S_OK(inform)
        if update['RequestState'] == 'Accepted':
          update['RequestState'] = 'Tech OK'
        if not update['RequestState'] in ['Tech OK', 'Rejected']:
          self.lock.release()
          _msgTuple = (requestState, update['RequestState'])
          return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
# AZ: removed from logic
#        if update['RequestState'] == 'Tech OK' and not update.get('ProID',old['ProID']):
#          self.lock.release()
#          return S_ERROR("Registered processing pass is required to sign for Tech OK")
    elif requestState in ['PPG OK', 'On-hold']:
      for x in update:
        if x not in ['RequestState', 'Comments', 'Inform', 'ProPath', 'ProID', 'ProDetail', 'Extra']:
          self.lock.release()
          return S_ERROR("%s can't be modified during Tech signing" % x)
      if 'RequestState' not in update:
        return S_OK(inform)
      if update['RequestState'] == 'Tech OK':
        update['RequestState'] = 'Accepted'
      if not update['RequestState'] in ['Accepted', 'Rejected', 'On-hold']:
        self.lock.release()
        _msgTuple = (requestState, update['RequestState'])
        return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
# AZ: removed from logic
#      if update['RequestState'] == 'Accepted' and not update.get('ProID',old['ProID']):
#        self.lock.release()
#        return S_ERROR("Registered processing pass is required to sign for Tech OK")
    elif requestState == 'Tech OK':
      for x in update:
        if x not in ['RequestState', 'RequestWG', 'Comments', 'Inform', 'RequestPriority', 'Extra']:
          self.lock.release()
          return S_ERROR("%s can't be modified during PPG signing" % x)
      if 'RequestState' not in update:
        return S_OK(inform)
      if update['RequestState'] == 'PPG OK':
        update['RequestState'] = 'Accepted'
      if not update['RequestState'] in ['Accepted', 'Rejected']:
        self.lock.release()
        _msgTuple = (requestState, update['RequestState'])
        return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
    elif requestState in ['Accepted', 'Active', 'Completed']:
      for x in update:
        if x not in ['RequestState', 'Comments', 'Inform', 'ProPath', 'ProID', 'ProDetail', 'Extra']:
          self.lock.release()
          return S_ERROR("%s can't be modified during the progress" % x)
      if 'RequestState' not in update:
        return S_OK(inform)
      if requestState == 'Accepted':
        if not update['RequestState'] in ['Active', 'Cancelled', 'PPG OK']:
          self.lock.release()
          _msgTuple = (requestState, update['RequestState'])
          return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
      elif requestState == 'Active':
        if (update['RequestState'] not in ['Done', 'Cancelled', 'Completed', 'Accepted']) or\
                (update['RequestState'] != 'Accepted' and creds['Group'] != 'lhcb_prmgr'):
          self.lock.release()
          _msgTuple = (requestState, update['RequestState'])
          return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
      else:
        if not update['RequestState'] in ['Active', 'Done', 'Cancelled']:
          self.lock.release()
          _msgTuple = (requestState, update['RequestState'])
          return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
    inform['state'] = update['RequestState']
    inform['rec'].update(update)
    return S_OK(inform)

  def updateProductionRequest(self, requestID, requestDict, creds):
    """Update existing production request In states other than New only state
    and comments are changable.

    TODO: RequestPDG change in ??? state
          Protect fields in subrequests
    """
    fdict = dict.fromkeys(self.requestFields[4:-7], None)
    rec = {}
    for x in requestDict:
      if x in fdict:
        if requestDict[x]:
          rec[x] = requestDict[x]  # set only known fields
        else:
          rec[x] = None  # to be more deterministic...

    self.lock.acquire()  # transaction begin ?? may be after connection ??
    result = self._getConnection()
    if not result['OK']:
      self.lock.release()
      return S_ERROR('Failed to get connection to MySQL: ' + result['Message'])
    connection = result['Value']

    fields = ','.join(['t.' + x for x in self.requestFields[:-7]])
    req = "SELECT %s " % fields
    req += "FROM ProductionRequests as t "
    req += "WHERE t.RequestID=%s" % requestID
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    if not result['Value']:
      self.lock.release()
      return S_ERROR('The request is no longer exist')

    old = dict(zip(self.requestFields[:-7], result['Value'][0]))

    update = {}     # Decide what to update (and if that is required)
    for x in rec:
      if x in ('ProDetail', 'SimCondDetail', 'Extra'):
        if rec[x] == old[x]:
          continue
        try:
          recx = json.loads(rec[x])
          oldx = json.loads(old[x])
          if recx == oldx:
            continue
        except TypeError:
          # This happens if, for example, oldx is None (meaning there was not prodetail, while now there is).
          # Which means that now we can update
          pass
      elif str(rec[x]) == str(old[x]):
        continue

      if x == 'RetentionRate' and float(rec[x]) == old[x]:
        continue
      if x == 'ProDetail':
        result = self.__checkIOTypes(requestDict)
        if not result['OK']:
          self.lock.release()
          return result
      update[x] = rec[x]

    if len(update) == 0:
      self.lock.release()
      return S_OK(requestID)  # nothing to update

    if 'NumberOfEvents' in update:  # Update RealNumberOfEvents if specified
      num = 0
      try:
        num = int(rec['NumberOfEvents'])
        if num < 0:
          num = 0
      except ValueError:
        pass
      except TypeError:
        gLogger.warn("NumberOfEvents not defined for %s" % requestID)
      update['RealNumberOfEvents'] = str(num)

    if 'Comments' in update:
      update['Comments'] = self.__prefixComments(update['Comments'],
                                                 old['Comments'], creds['User'])

    result = self.__checkUpdate(update, old, creds, connection)
    if not result['OK']:
      return result
    inform = result['Value']

    # we have to escape values... tricky way
    # in addition we can not escape the datetime.date values. so if they are being updated,
    # they will be added to this list later
    recl_fields = [column for column in update if column not in self.dateColumns]
    recl = [update[x] for x in recl_fields]
    result = self._fixedEscapeValues(recl)
    if not result['OK']:
      self.lock.release()
      return result
    updateValues = result['Value']
    for dateValue in self.dateColumns:
      # we are going to check if we have column which type is date
      # the order is very important that's why we use recl_fields
      if dateValue in self.dateColumns and requestDict.get(dateValue):
        recl_fields.append(dateValue)
        updateValues.append("STR_TO_DATE('%s','%s')" % (requestDict.get(dateValue), self.dateFormat))
    updates = ','.join([x + '=' + y for x, y in zip(recl_fields, updateValues)])

    req = "UPDATE ProductionRequests "
    req += "SET %s " % updates
    req += "WHERE RequestID=%s" % requestID
    result = self._update(req, connection)
    if not result['OK']:
      self.lock.release()
      return result

    if 'RequestState' in update:
      result = self._update("INSERT INTO RequestHistory (" +
                            ','.join(self.historyFields[:-1]) +
                            ") VALUES ( %s,'%s','%s')" %
                            (requestID, str(update['RequestState']),
                             str(creds['User'])), connection)
      if not result['OK']:
        gLogger.error(result['Message'])

    result = S_OK()
    gLogger.info(str(update))

    if result['OK']:
      self.lock.release()

    informPeople(**inform)
    return S_OK(requestID)

  def __getSubrequestsList(self, iD, master, connection):
    ''' Return list of all subrequests for this request
        NOTE: it does self.lock.release() in case of errors
    '''
    result = self._query("SELECT RequestID " +
                         "FROM ProductionRequests " +
                         "WHERE ParentID=%s and MasterID=%s" % (iD, master),
                         connection)
    if not result['OK']:
      self.lock.release()
      return result
    sr = []
    for x in result['Value']:
      sr.append(x[0])
      result = self.__getSubrequestsList(x[0], master, connection)
      if not result['OK']:
        self.lock.release()
        return result
      sr += result['Value']
    return S_OK(sr)

  def deleteProductionRequest(self, requestID, creds):
    """Delete existing production.

    Subrequests are deleted. Substructure is moved up in the tree.
    Available is New and Rejected states only
    """
    try:
      requestID = int(requestID)
    except ValueError:
      return S_ERROR('RequestID is not a number')
    self.lock.acquire()  # transaction begin ?? may be after connection ??
    result = self._getConnection()
    if not result['OK']:
      self.lock.release()
      return S_ERROR('Failed to get connection to MySQL: ' + result['Message'])
    connection = result['Value']

    result = self.__getStateAndAuthor(requestID, connection)
    if not result['OK']:
      return result
    requestState, requestAuthor, _requestInform, isModel = result['Value']
    if creds['Group'] not in ['diracAdmin', 'lhcb_admin']:
      if requestAuthor != creds['User'] and not (isModel and creds['Group'] == 'lhcb_tech'):
        self.lock.release()
        gLogger.error("%s can't remove %s request" % (creds['User'], requestAuthor))
        return S_ERROR('Only author can remove a request')
      if requestState != 'New' and requestState != 'Rejected':
        self.lock.release()
        return S_ERROR('Can not remove request in processing')
    result = self.__getRequestInfo(requestID, connection)
    if not result['OK']:
      return result
    pinfo = result['Value']
    parentID = pinfo['ParentID']
    masterID = pinfo['MasterID']

    upperID = parentID
    if not upperID:
      upperID = "NULL"

    # delete subrequests
    req = ''
    if not masterID:  # this request is a master
      # delete history
      result = self._update("DELETE FROM RequestHistory " +
                            "WHERE RequestID=%s" % requestID)
      if not result['OK']:
        self.lock.release()
        return result
      # delete tracking
      result = self._update("DELETE FROM ProductionProgress " +
                            "WHERE RequestID IN " +
                            "(SELECT RequestID FROM ProductionRequests " +
                            "WHERE RequestID=%s OR MasterID=%s" %
                            (requestID, requestID) + ")", connection)
      if not result['OK']:
        self.lock.release()
        return result
      req = "DELETE FROM ProductionRequests "
      req += "WHERE MasterID=%s" % requestID
    else:  # this request is subrequest
      result = self.__getSubrequestsList(requestID, masterID, connection)
      if not result['OK']:
        return result
      rlist = result['Value']
      # delete tracking
      result = self._update("DELETE FROM ProductionProgress " +
                            "WHERE RequestID IN (%s)" %
                            ','.join([str(x) for x in rlist] + [str(requestID)]))
      if not result['OK']:
        self.lock.release()
        return result
      if len(rlist):
        req = "DELETE FROM ProductionRequests "
        req += "WHERE RequestID in (%s)" % ','.join([str(x) for x in rlist])
    if req:
      result = self._update(req, connection)
      if not result['OK']:
        self.lock.release()
        return result

    # move substructure
    req = "UPDATE ProductionRequests SET ParentID=%s " % upperID
    req += "WHERE ParentID=%s" % requestID
    result = self._update(req, connection)
    if not result['OK']:
      self.lock.release()
      return result

    # finally delete us
    req = "DELETE FROM ProductionRequests "
    req += "WHERE RequestID=%s" % requestID
    result = self._update(req, connection)

    self.lock.release()

    if not result['OK']:
      return result
    return S_OK(requestID)

  def __getRequest(self, requestID, connection):
    """retrive complete request record.

    NOTE: unlock in case of errors
    """
    fields = ','.join(['t.' + x for x in self.requestFields[:-7]])
    req = "SELECT %s " % fields
    req += "FROM ProductionRequests as t "
    req += "WHERE t.RequestID=%s" % requestID
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    if not result['Value']:
      self.lock.release()
      return S_ERROR('The request is no longer exist')
    rec = dict(zip(self.requestFields[:-7], result['Value'][0]))
    return S_OK(rec)

  @staticmethod
  def __clearProcessingPass(rec):
    """clear processing pass section."""
    rec['ProID'] = None
    nd = {}
    rec['ProDetail'] = json.dumps(nd)

  def __duplicateDeep(self, requestID, masterID, parentID, creds, connection, clearpp):
    """recurcive duplication function.

    NOTE: unlock in case of errors
    """

    result = self.__getRequest(requestID, connection)
    if not result['OK']:
      self.lock.release()
      return result
    rec = result['Value']
    if clearpp:
      self.__clearProcessingPass(rec)
    rec['IsModel'] = 0

    if masterID and not rec['MasterID']:
      return S_OK("")  # substructured request

    if rec['ParentID']:
      rec['ParentID'] = parentID
    if rec['MasterID']:
      rec['MasterID'] = masterID
    if rec['RequestAuthor']:
      rec['RequestAuthor'] = creds['User']
    if rec['RequestState']:
      rec['RequestState'] = 'New'

    # Clear RealNumberOfEvents if required
    try:
      num = 0
      num = int(rec['NumberOfEvents'])
      if num < 0:
        num = 0
    except ValueError:
      pass
    except TypeError:
      gLogger.warn("NumberOfEvents is not defined for %s" % requestID)
    rec['RealNumberOfEvents'] = str(num)

    recl = [rec[x] for x in self.requestFields[1:-7]]
    escapeIndex = self.requestFields[1:-7].index('FastSimulationType')
    result = self._fixedEscapeValues(recl, escapeIndex)
    if not result['OK']:
      self.lock.release()
      return result
    recls = result['Value']

    req = "INSERT INTO ProductionRequests ( " + ','.join(self.requestFields[1:-7])
    req += " ) VALUES ( %s );" % ','.join(recls)
    result = self._update(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    req = "SELECT LAST_INSERT_ID();"
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    newRequestID = int(result['Value'][0][0])

    # Update history for masters. Errors are not reported back to the user.
    if rec['RequestState']:
      result = self._update("INSERT INTO RequestHistory (" +
                            ','.join(self.historyFields[:-1]) +
                            ") VALUES ( %s,'%s','%s')" %
                            (newRequestID, str(rec['RequestState']),
                             str(rec['RequestAuthor'])), connection)
      if not result['OK']:
        gLogger.error(result['Message'])

    # now for subrequests
    if not masterID:
      masterID = newRequestID
    parentID = newRequestID

    req = "SELECT RequestID "
    req += "FROM ProductionRequests as t "
    req += "WHERE t.ParentID=%s" % requestID
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    for chID in [row[0] for row in result['Value']]:
      result = self.__duplicateDeep(chID, masterID, parentID, creds, connection, False)
      if not result['OK']:
        return result

    return S_OK(int(newRequestID))

  def duplicateProductionRequest(self, requestID, creds, clearpp):
    """Duplicate production request with all it's subrequests (but without
    substructure).

    If that is subrequest, master must be in New state and user must be
    the author. If clearpp is set, all details in the Processing pass
    (of the master) are cleaned.
    """
    try:
      requestID = int(requestID)
    except ValueError:
      return S_ERROR('RequestID is not a number')
    self.lock.acquire()  # transaction begin ?? may be after connection ??
    result = self._getConnection()
    if not result['OK']:
      self.lock.release()
      return S_ERROR('Failed to get connection to MySQL: ' + result['Message'])
    connection = result['Value']

    result = self.__getRequestInfo(requestID, connection)
    if not result['OK']:
      return result
    pinfo = result['Value']
    parentID = pinfo['ParentID']
    masterID = pinfo['MasterID']

    if masterID:
      clearpp = False
      result = self.__getStateAndAuthor(requestID, connection)
      if not result['OK']:
        return result
      requestState, requestAuthor, _requestInform, isModel = result['Value']
      if requestState != 'New' or (requestAuthor != creds['User'] and not (isModel and creds['Group'] == 'lhcb_tech')):
        self.lock.release()
        return S_ERROR('Can not duplicate subrequest of request in progress')

    result = self.__duplicateDeep(requestID, masterID, parentID, creds, connection, clearpp)
    if result['OK']:
      self.lock.release()
    return result

  @staticmethod
  def __checkAuthorizeSplit(requestState, creds):
    """Check that current user is allowed to split in specified state."""
    if creds['Group'] in ['diracAdmin', 'lhcb_admin']:
      return S_OK()
    if (requestState in ['Submitted', 'PPG OK', 'On-hold']) and creds['Group'] == 'lhcb_tech':
      return S_OK()
    if requestState in ['Accepted', 'Active', 'Completed'] and creds['Group'] == 'lhcb_prmgr':
      return S_OK()
    return S_ERROR('You are not allowed to split the request')

  def __moveChildDeep(self, requestID, masterID, setParent, connection):
    """Update parent for this request if setParent is True and update master
    for this and all subrequests."""
    if setParent:
      updates = "ParentID=%s,MasterID=%s" % (str(masterID), str(masterID))
    req = "UPDATE ProductionRequests "
    req += "SET %s " % updates
    req += "WHERE RequestID=%s" % requestID
    result = self._update(req, connection)
    if not result['OK']:
      return result
    req = "SELECT RequestID,MasterID "
    req += "FROM ProductionRequests as t "
    req += "WHERE t.ParentID=%s" % requestID
    result = self._query(req, connection)
    if not result['OK']:
      return result
    for ch in result['Value']:
      if ch[1]:
        ret = self.__moveChildDeep(ch[0], masterID, False, connection)
        if not ret['OK']:
          gLogger.error("_moveChildDeep: can not move to %s: %s" % (str(masterID), ret['Message']))
        # !!! Failing that will leave both requests inconsistant !!!
        # But since we have only one subrequest level now, it will never happened
    return S_OK()

  def splitProductionRequest(self, requestID, splitlist, creds):
    """Fully duplicate master production request with its history and
    reassociate first level subrequests from splitlist (with there subrequest
    structure).

    Substructures can not be moved. Only experts in appropriate request
    state can request the split.
    """
    try:
      requestID = int(requestID)
    except ValueError:
      return S_ERROR('RequestID is not a number')
    if not splitlist:
      return S_ERROR('Split list is empty')
    isplitlist = []
    try:
      isplitlist = [int(x) for x in splitlist]
    except ValueError:
      return S_ERROR('RequestID in split list is not a number')

    self.lock.acquire()  # transaction begin ?? may be after connection ??
    result = self._getConnection()
    if not result['OK']:
      self.lock.release()
      return S_ERROR('Failed to get connection to MySQL: ' + result['Message'])
    connection = result['Value']

    result = self.__getRequest(requestID, connection)
    if not result['OK']:
      self.lock.release()
      return result
    rec = result['Value']

    # Check that operation is valid
    if rec['MasterID']:
      self.lock.release()
      return S_ERROR('Can not duplicate subrequest of request in progress')
    result = self.__checkAuthorizeSplit(rec['RequestState'], creds)
    if not result['OK']:
      self.lock.release()
      return result
    req = "SELECT RequestID,MasterID "
    req += "FROM ProductionRequests as t "
    req += "WHERE t.ParentID=%s" % requestID
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    fisplitlist = []
    keeplist = []
    for ch in result['Value']:
      if ch[0] in isplitlist and ch[1] == requestID:
        fisplitlist.append(int(ch[0]))
      elif ch[1]:
        keeplist.append(int(ch[0]))
    if len(isplitlist) != len(fisplitlist):
      self.lock.release()
      return S_ERROR('Requested for spliting subrequests are no longer exist')
    if not keeplist:
      self.lock.release()
      return S_ERROR('You have to keep at least one subrequest')

    # Now copy the master
    recl = [rec[x] for x in self.requestFields[1:-7]]
    result = self._fixedEscapeValues(recl, 24)
    if not result['OK']:
      self.lock.release()
      return result
    recls = result['Value']
    req = "INSERT INTO ProductionRequests ( " + ','.join(self.requestFields[1:-7])
    req += " ) VALUES ( %s );" % ','.join(recls)
    result = self._update(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    req = "SELECT LAST_INSERT_ID();"
    result = self._query(req, connection)
    if not result['OK']:
      self.lock.release()
      return result
    newRequestID = int(result['Value'][0][0])

    # Move subrequests (!! Errors are not fatal !!)
    rsplitlist = []
    for x in fisplitlist:
      if self.__moveChildDeep(x, newRequestID, True, connection)['OK']:
        rsplitlist.append(x)
    # If nothing could be moved, remove the master
    if not rsplitlist:
      req = "DELETE FROM ProductionRequests "
      req += "WHERE RequestID=%s" % str(newRequestID)
      result = self._update(req, connection)
      self.lock.release()
      return S_ERROR('Could not move subrequests')

    # Copy the history (failures are not fatal since
    # it is hard to revert the previous changes...)
    req = "SELECT " + ','.join(self.historyFields) + " FROM RequestHistory WHERE RequestID=%s " % requestID
    req += "ORDER BY TimeStamp"
    result = self._query(req, connection)
    if not result['OK']:
      gLogger.error("SplitProductionRequest: can not get history for %s: %s" % (str(requestID),
                                                                                result['Message']))
    else:
      for x in result['Value']:
        x = list(x)
        x[0] = newRequestID
        ret = self._update("INSERT INTO RequestHistory (" +
                           ','.join(self.historyFields) +
                           ") VALUES ( %s,'%s','%s','%s')" % tuple([str(y) for y in x]), connection)
      if not ret['OK']:
        gLogger.error("SplitProductionRequest: add history fail: %s", ['Message'])

    self.lock.release()
    return S_OK(newRequestID)

  progressFields = ['ProductionID', 'RequestID', 'Used', 'BkEvents']

  def getProductionProgress(self, requestID):
    """return the list of associated productions requestID must be Long and
    already checked."""
    req = "SELECT * FROM ProductionProgress WHERE RequestID=%s" % requestID
    result = self._query(req)
    if not result['OK']:
      return result

    rows = [dict(zip(self.progressFields, row)) for row in result['Value']]
    total = len(rows)
    return S_OK({'Rows': rows, 'Total': total})

  def addProductionToRequest(self, pdict):
    """Associate production to request.

    Existence of request is checked first.
    TODO: check requestState
    """
    try:
      for x in self.progressFields:
        pdict[x] = int(pdict[x])
    except ValueError:
      return S_ERROR('Bad parameters')

    self.lock.acquire()  # transaction begin ?? may be after connection ??
    result = self._getConnection()
    if not result['OK']:
      self.lock.release()
      return S_ERROR('Failed to get connection to MySQL: ' + result['Message'])
    connection = result['Value']

    result = self.__getStateAndAuthor(pdict['RequestID'], connection)
    if not result['OK']:
      self.lock.release()
      return result
    # requestState, requestAuthor, requestInform = result['Value']

    req = "INSERT INTO ProductionProgress ( "
    req += ','.join(self.progressFields)
    req += " ) VALUES ( "
    req += ','.join([str(pdict[x]) for x in self.progressFields])
    req += " )"
    result = self._update(req, connection)
    self.lock.release()
    if not result['OK']:
      return result
    return S_OK(pdict['ProductionID'])

  def removeProductionFromRequest(self, productionID):
    """Deassociate production."""
    req = "DELETE FROM ProductionProgress "
    req += "WHERE ProductionID=%s" % str(productionID)
    result = self._update(req)
    if not result['OK']:
      return result
    return S_OK(productionID)

  def useProductionForRequest(self, productionID, used):
    """Deassociate production."""
    used = int(used)
    req = "UPDATE ProductionProgress "
    req += "SET Used=%s " % str(used)
    req += "WHERE ProductionID=%s" % str(productionID)
    result = self._update(req)
    if not result['OK']:
      return result
    return S_OK(productionID)

  def getRequestHistory(self, requestID):
    """return the list of state changes for the requests requestID must be Long
    and already checked."""
    req = "SELECT " + ','.join(self.historyFields)
    req += " FROM RequestHistory WHERE RequestID=%s " % requestID
    req += "ORDER BY TimeStamp"
    result = self._query(req)
    if not result['OK']:
      return result

    rows = [dict(zip(self.historyFields, row)) for row in result['Value']]
    total = len(rows)
    return S_OK({'Rows': rows, 'Total': total})

  def getTrackedProductions(self):
    """return a list of all productions associated with requests in 'Active' or
    'Completed' state."""
    req1 = "SELECT RequestID FROM ProductionRequests WHERE RequestState in ('Active','Completed')"
    req2 = "SELECT RequestID FROM ProductionRequests WHERE RequestState in ('Active','Completed')"
    req2 += " OR MasterID in (%s)" % req1
    req = "SELECT ProductionID FROM ProductionProgress WHERE RequestID "
    req += "in (%s)" % req2

    result = self._query(req)
    if not result['OK']:
      return result
    values = [row[0] for row in result['Value']]
    return S_OK(values)

  def updateTrackedProductions(self, update):
    """update tracked productions."""
    # check parameters
    try:
      for x in update:
        x['ProductionID'] = int(x['ProductionID'])
        x['BkEvents'] = int(x['BkEvents'])
    except ValueError:
      return S_ERROR('Bad parameters')
    except TypeError:
      return S_ERROR('Parameters not defined')

    for x in update:
      result = self._update("UPDATE ProductionProgress " +
                            "SET BkEvents=%s " % x['BkEvents'] +
                            "WHERE ProductionID=%s" % x['ProductionID'])
      if not result['OK']:
        gLogger.info('Problem in updating progress. Not fatal: %s' % result['Message'])
    return S_OK('')

  def __trackedInputSQL(self, fields):
    req = "SELECT %s " % fields
    req += "FROM ProductionRequests as t WHERE"
    req += ' t.RequestState in ("Active","Completed")'
    req += ' AND NumberOfEvents<0 '
    return self._query(req)

  def getTrackedInput(self):
    """return a list of all requests with dynamic input in 'Active' or
    'Completed' states."""

    fields = ','.join(['t.' + x for x in self.requestFields[:-7]])
    result = self.__trackedInputSQL(fields)
    if not result['OK']:
      return result
    rec = []
    for x in result['Value']:
      res = dict(zip(self.requestFields[:-7], x))
      if res['SimCondDetail']:
        res.update(json.loads(res['SimCondDetail']))
      else:
        continue
      del res['SimCondDetail']
      try:
        num = int(res['RealNumberOfEvents'])
      except ValueError:
        num = 0
      except TypeError:
        num = 0
      if num > 0:
        continue
      rec.append(res)
    return S_OK(rec)

  def updateTrackedInput(self, update):
    """update real number of input events."""
    # check parameters
    try:
      for x in update:
        x['RequestID'] = int(x['RequestID'])
        x['RealNumberOfEvents'] = int(x['RealNumberOfEvents'])
    except ValueError:
      return S_ERROR('Bad parameters')
    except TypeError:
      return S_ERROR('Parameters not defined')
    result = self.__trackedInputSQL('RequestID')
    if not result['OK']:
      return result
    allowed = dict.fromkeys([x[0] for x in result['Value']], None)
    skiped = []
    for x in update:
      if not x['RequestID'] in allowed:
        skiped.append(x['RequestID'])
        continue
      req = "UPDATE ProductionRequests "
      req += "SET RealNumberOfEvents=%s " % str(x['RealNumberOfEvents'])
      req += "WHERE RequestID=%s" % str(x['RequestID'])
      result = self._update(req)
      if not result['OK']:
        skiped.append(x['RequestID'])
    if skiped:
      return S_ERROR("updateTrackedInput has skiped requests %s" %
                     ','.join(skiped))
    return S_OK('')

  def getProductionList(self, requestID):
    """return a list of all productions associated with the request or any its
    subrequest."""
    req1 = "SELECT RequestID FROM ProductionRequests WHERE MasterID=%s" % requestID
    req = "SELECT ProductionID FROM ProductionProgress WHERE RequestID "
    req += "in (%s) OR RequestID=%s" % (req1, requestID)

    result = self._query(req)
    if not result['OK']:
      return result

    gLogger.info(result['Value'])

    values = [row[0] for row in result['Value']]
    return S_OK(values)

  def getAllSubRequestSummary(self, status='', rType=''):
    """return a dictionary containing a summary for each subrequest."""
    req = "SELECT RequestID,ParentID,RequestType,RequestState,NumberOfEvents FROM ProductionRequests"
    if status and rType:
      req = "%s WHERE RequestState = '%s' AND RequestType = '%s'" % (req, status, rType)
    elif status:
      req = "%s WHERE RequestState = '%s'" % (req, status)
    elif rType:
      req = "%s WHERE RequestType = '%s'" % (req, rType)
    res = self._query(req)
    if not res['OK']:
      return res
    sRequestInfo = {}
    for sRequestID, parentID, tReq, status, reqEvents in res['Value']:
      if not parentID:
        parent = 0
      sRequestInfo[sRequestID] = {'Master': parent,
                                  'RequestType': tReq,
                                  'Status': status,
                                  'ReqEvents': reqEvents}
    return S_OK(sRequestInfo)

  def getAllProductionProgress(self):
    """return a dictionary containing for each requestID the active productions
    and the number of events."""
    req = "SELECT RequestID, ProductionID, Used, BkEvents FROM ProductionProgress;"
    res = self._query(req)
    if not res['OK']:
      return res
    sRequestInfo = {}
    for sRequestID, prodID, used, events in res['Value']:
      if sRequestID not in sRequestInfo:
        sRequestInfo[sRequestID] = {}
      sRequestInfo[sRequestID][prodID] = {'Used': used, 'Events': events}
    return S_OK(sRequestInfo)

  def getFilterOptions(self):
    """Return the dictionary with possible values for filter."""
    opts = {}
    for key, value in [('State', 'RequestState'),
                       ('Type', 'RequestType'),
                       ('Author', 'RequestAuthor'),
                       ('EType', 'EventType'),
                       ('WG', 'RequestWG')]:

      req = "SELECT DISTINCT %s FROM ProductionRequests " % value
      req += "WHERE %s IS NOT NULL " % value
      req += "ORDER BY %s" % value
      res = self._query(req)
      if not res['OK']:
        return res
      opts[key] = [row[0] for row in res['Value']]
    return S_OK(opts)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""GangaDataFile is a utility to create a Data file, to be used by ganga.

Givn input files, it will create something like::

  from Gaudi.Configuration import *
  from GaudiConf import IOHelper
  IOHelper("ROOT").inputFiles([
      "LFN:foo",
      "LFN:bar"
  ], clear=True)

  FileCatalog().Catalogs = ["xmlcatalog_file:pool_xml_catalog.xml"]
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import six

from DIRAC import gLogger

from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient


class GangaDataFile(object):
  """Creates ganga data file."""

  def __init__(self, fileName='data.py', xmlcatalog_file='pool_xml_catalog.xml', log=None):
    """initialize."""
    if not log:
      self.log = gLogger.getSubLogger('GangaDataFile')
    else:
      self.log = log

    self.fileName = fileName
    self.xmlcatalog_file = xmlcatalog_file

    try:
      os.remove(self.fileName)
    except OSError:
      pass

    self.log.info('Creating Ganga data file %s from scratch' % self.fileName)

  ################################################################################

  def generateDataFile(self, lfns, persistency=None):
    """generate the data file."""
    if isinstance(lfns, six.string_types) and lfns:
      lfns = [lfns]
    elif not isinstance(lfns, list):
      self.log.error('Was expecting a list')
      raise TypeError('Expected List')
    if not len(lfns):
      self.log.warn('No file generated: was expecting a non-empty list')
      raise ValueError('list empty')

    try:
      persistency = persistency.upper()
    except AttributeError:
      pass

    # Create a fake LFN->PFN dictionary to give the persistency
    fakePfns = dict.fromkeys(lfns, {'pfntype': persistency})
    script = LHCB_BKKDBClient(welcome=False).writeJobOptions(lfns,
                                                             optionsFile=self.fileName,
                                                             catalog=self.xmlcatalog_file,
                                                             savePfn=fakePfns)
    self.log.info('Created Ganga data file %s' % self.fileName)

    return script



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Bookkeeping Reporting module (just prepare the files, do not send them
(which is done in the uploadOutput)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from collections import defaultdict
import os
import re
import shlex
import socket
import time
import psutil
from xml.dom.minidom import Document, DocumentType

from DIRAC import gLogger, S_OK, S_ERROR, gConfig
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Resources.Catalog.PoolXMLFile import getGUID
from DIRAC.Workflow.Utilities.Utils import getStepCPUTimes

import LHCbDIRAC
from LHCbDIRAC.Resources.Catalog.PoolXMLFile import getOutputType
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.ProductionData import constructProductionLFNs
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary, XMLSummaryError
from LHCbDIRAC.Core.Utilities.XMLTreeParser import addChildNode


class BookkeepingReport(ModuleBase):
  """BookkeepingReport class."""

  def __init__(self, bkClient=None, dm=None):
    """Usual c'tor."""

    self.log = gLogger.getSubLogger("BookkeepingReport")

    super(BookkeepingReport, self).__init__(self.log, bkClientIn=bkClient, dm=dm)

    self.version = __RCSID__

    self.simDescription = 'NoSimConditions'
    self.eventType = ''
    self.poolXMLCatName = ''
    self.stepInputData = []
    self.firstStepInput = ''
    self.jobType = ''
    self.stepOutputs = []
    self.histogram = False
    self.xf_o = None

    self.ldate = None
    self.ltime = None
    self.ldatestart = None
    self.ltimestart = None

  ################################################################################

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_number=None, step_id=None, saveOnFile=True):
    """Usual executor."""
    try:

      super(BookkeepingReport, self).execute(self.version, production_id, prod_job_id, wms_job_id,
                                             workflowStatus, stepStatus,
                                             wf_commons, step_commons, step_number, step_id)

      if not self._checkWFAndStepStatus():
        return S_OK()

      bkLFNs, logFilePath = self._resolveInputVariables()

      doc = self.__makeBookkeepingXML(bkLFNs, logFilePath)

      if saveOnFile:
        bfilename = 'bookkeeping_' + self.step_id + '.xml'
        with open(bfilename, 'wb') as bfile:
          bfile.write(doc)
      else:
        print(doc)

      return S_OK()

    except Exception as e:  # pylint:disable=broad-except
      self.log.exception("Failure in BookkeepingReport execute module", lException=e)
      self.setApplicationStatus(repr(e))
      return S_ERROR(str(e))

    finally:
      super(BookkeepingReport, self).finalize(self.version)

################################################################################
# AUXILIAR FUNCTIONS
################################################################################

  def _resolveInputVariables(self):
    """By convention the module parameters are resolved here."""

    super(BookkeepingReport, self)._resolveInputVariables()
    super(BookkeepingReport, self)._resolveInputStep()

    self.stepOutputs, _sot, _hist = self._determineOutputs()

    # # VARS FROM WORKFLOW_COMMONS ##

    if 'outputList' in self.workflow_commons:
      for outputItem in self.stepOutputs:
        if outputItem not in self.workflow_commons['outputList']:
          self.workflow_commons['outputList'].append(outputItem)
    else:
      self.workflow_commons['outputList'] = self.stepOutputs

    if 'BookkeepingLFNs' in self.workflow_commons and \
        'LogFilePath' in self.workflow_commons and \
            'ProductionOutputData' in self.workflow_commons:

      logFilePath = self.workflow_commons['LogFilePath']
      bkLFNs = self.workflow_commons['BookkeepingLFNs']

      if not isinstance(bkLFNs, list):
        bkLFNs = [i.strip() for i in bkLFNs.split(';')]

    else:
      self.log.info('LogFilePath / BookkeepingLFNs parameters not found, creating on the fly')
      result = constructProductionLFNs(self.workflow_commons, self.bkClient)
      if not result['OK']:
        self.log.error('Could not create production LFNs', result['Message'])
        raise ValueError('Could not create production LFNs')

      bkLFNs = result['Value']['BookkeepingLFNs']
      logFilePath = result['Value']['LogFilePath'][0]

    self.ldate = time.strftime("%Y-%m-%d", time.gmtime(time.time()))
    self.ltime = time.strftime("%H:%M:%S", time.gmtime(time.time()))

    if 'StartTime' in self.step_commons:
      startTime = self.step_commons['StartTime']
      self.ldatestart = time.strftime("%Y-%m-%d", time.gmtime(startTime))
      self.ltimestart = time.strftime("%H:%M:%S", time.gmtime(startTime))

    try:
      self.xf_o = self.step_commons['XMLSummary_o']
    except KeyError:
      self.log.warn('XML Summary object not found, will try to create it (again?)')
      try:
        xmlSummaryFile = self.step_commons['XMLSummary']
      except KeyError:
        self.log.warn('XML Summary file name not found, will try to guess it')
        xmlSummaryFile = 'summary%s_%s_%s_%s.xml' % (self.applicationName,
                                                     self.production_id,
                                                     self.prod_job_id,
                                                     self.step_number)
        self.log.warn('Trying %s' % xmlSummaryFile)
        if xmlSummaryFile not in os.listdir('.'):
          self.log.warn('XML Summary file %s not found, will try to guess a second time' % xmlSummaryFile)
          xmlSummaryFile = 'summary%s_%s.xml' % (self.applicationName,
                                                 self.step_id)
          self.log.warn('Trying %s' % xmlSummaryFile)
          if xmlSummaryFile not in os.listdir('.'):
            self.log.warn('XML Summary file %s not found, will try to guess a third and last time' % xmlSummaryFile)
            xmlSummaryFile = 'summary%s_%s.xml' % (self.applicationName,
                                                   self.step_number)
            self.log.warn('Trying %s' % xmlSummaryFile)
      try:
        self.xf_o = XMLSummary(xmlSummaryFile)
      except XMLSummaryError as e:
        self.log.warn('No XML summary available',
                      '%s' % repr(e))
        self.xf_o = None

    return bkLFNs, logFilePath

################################################################################
################################################################################

  def __makeBookkeepingXML(self, bkLFNs, logFilePath):
    """Bookkeeping xml looks like this::

        <Job ConfigName="" ConfigVersion="" Date="" Time="">
          <TypedParameter Name="" Type="" Value=""/>
          ...
          <InputFile Name=""/>
          ...
          <OutputFile Name="" TypeName="" TypeVersion="">
            <Parameter Name="" Value=""/>
            ...
            <Replica Location="" Name=""/>
            ....
          </OutputFile>
          ...
          <SimulationCondition>
            <Parameter Name="" Value=""/>
          </SimulationCondition>
        </Job>

    """
    # Generate XML document
    doc = Document()
    docType = DocumentType("Job")
    docType.systemId = "book.dtd"
    doc.appendChild(docType)

    # Generate JobNode
    doc, jobNode = self.__generateJobNode(doc)
    # Generate TypedParams
    jobNode = self.__generateTypedParams(jobNode)
    # Generate InputFiles
    jobNode = self.__generateInputFiles(jobNode, bkLFNs)
    # Generate OutputFiles
    jobNode = self.__generateOutputFiles(jobNode, bkLFNs, logFilePath)
    # Generate SimulationConditions
    jobNode = self.__generateSimulationCondition(jobNode)

    prettyXMLDoc = doc.toprettyxml(indent="    ", encoding="ISO-8859-1")

    # horrible, necessary hack!
    prettyXMLDoc = prettyXMLDoc.replace(b'\'book.dtd\'', b'\"book.dtd\"')

    return prettyXMLDoc

################################################################################

  def __generateJobNode(self, doc):
    """Node looks like.this::

        <Job ConfigName="" ConfigVersion="" Date="" Time="">
    """

    # Get the Config name from the environment if any
    if 'configName' in self.workflow_commons:
      configName = self.workflow_commons['configName']
      configVersion = self.workflow_commons['configVersion']
    else:
      configName = self.applicationName
      configVersion = self.applicationVersion

    jobAttributes = (configName, configVersion, self.ldate, self.ltime)

    return addChildNode(doc, "Job", 1, jobAttributes)

################################################################################

  def __generateTypedParams(self, jobNode):
    """TypedParameter looks like this::

        <TypedParameter Name="" Type="" Value="">

      List of possible TypedParameter names
      - CPUTIME
      - ExecTime
      - WNMODEL
      - WNMEMORY
      - WNCPUPOWER
      - WNCACHE
      - WNCPUHS06
      - WNMJFHS06
      - Production
      - DiracJobId
      - Name
      - JobStart
      - JobEnd
      - Location
      - JobType
      - WorkerNode
      - GeometryVersion
      - ProgramName
      - ProgramVersion
      - DiracVersion
      - FirstEventNumber
      - StatisticsRequested
      - NumberOfEvents
      - StepID
    """

    typedParams = []

    # Timing
    exectime, cputime = getStepCPUTimes(self.step_commons)

    typedParams.append(("CPUTIME", cputime))
    typedParams.append(("ExecTime", exectime))

    res = self.__getNodeInformation()
    nodeInfo = defaultdict(lambda: 'unknown')
    if res['OK']:
      nodeInfo.update(res['Value'])

    typedParams.append(("WNMODEL", nodeInfo["ModelName"]))
    typedParams.append(("WNCPUPOWER", nodeInfo["CPU(MHz)"]))
    typedParams.append(("WNCACHE", nodeInfo["CacheSize(kB)"]))

    host = os.environ.get("HOSTNAME", os.environ.get("HOST"))
    typedParams.append(("WorkerNode", host or nodeInfo['HostName']))

    try:
      memory = self.xf_o.memory
    except AttributeError:
      memory = nodeInfo["Memory(kB)"]

    typedParams.append(("WNMEMORY", memory))

    diracPower = gConfig.getValue("/LocalSite/CPUNormalizationFactor", "0")
    typedParams.append(("WNCPUHS06", diracPower))
    mjfPower = gConfig.getValue("/LocalSite/CPUScalingFactor", "0")
    # Trick to know that the value is obtained from MJF: # from diracPower
    if mjfPower != diracPower:
      typedParams.append(("WNMJFHS06", mjfPower))
    typedParams.append(("NumberOfProcessors", self.numberOfProcessors))
    typedParams.append(("Production", self.production_id))
    typedParams.append(("DiracJobId", str(self.jobID)))
    typedParams.append(("Name", self.step_id))
    typedParams.append(("JobStart", '%s %s' % (self.ldatestart, self.ltimestart)))
    typedParams.append(("JobEnd", '%s %s' % (self.ldate, self.ltime)))
    typedParams.append(("Location", self.siteName))
    typedParams.append(("JobType", self.jobType))

    if 'XMLDDDB_VERSION' in os.environ:
      typedParams.append(("GeometryVersion", os.environ["XMLDDDB_VERSION"]))

    typedParams.append(("ProgramName", self.applicationName))
    typedParams.append(("ProgramVersion", self.applicationVersion))

    # DIRAC version
    typedParams.append(("DiracVersion", LHCbDIRAC.__version__))

    typedParams.append(("FirstEventNumber", 1))

    typedParams.append(("StatisticsRequested", self.numberOfEvents))

    typedParams.append(("StepID", self.BKstepID))

    try:
      noOfEvents = self.xf_o.inputEventsTotal if self.xf_o.inputEventsTotal else self.xf_o.outputEventsTotal
    except AttributeError:
      # This happens iff the XML summary can't be created (e.g. for merging MDF files)
      res = self.bkClient.getFileMetadata(self.stepInputData)
      if not res['OK']:
        raise AttributeError("Can't get the BKK file metadata")
      noOfEvents = sum(fileMeta['EventStat'] for fileMeta in res['Value']['Successful'].values())

    typedParams.append(("NumberOfEvents", noOfEvents))

    # Add TypedParameters to the XML file
    for typedParam in typedParams:
      jobNode = addChildNode(jobNode, "TypedParameter", 0, typedParam)

    return jobNode

################################################################################

  def __generateInputFiles(self, jobNode, bkLFNs):
    """InputData looks like this::

         <InputFile Name=""/>
    """

    self.log.debug("Adding InputData: bkLFNs = %s" % bkLFNs)
    self.log.debug("Adding InputData: self.stepInputData = %s" % self.stepInputData)

    if self.stepInputData:
      intermediateInputs = False
      for inputname in self.stepInputData:
        for bkLFN in bkLFNs:
          if os.path.basename(bkLFN).lower() == os.path.basename(inputname).lower():
            # preserve the case
            inputF = os.path.join(os.path.dirname(os.path.normpath(bkLFN)),
                                  os.path.basename(inputname))
            jobNode = addChildNode(jobNode, "InputFile", 0, (inputF, ))
            intermediateInputs = True
        if not intermediateInputs:
          jobNode = addChildNode(jobNode, "InputFile", 0, (inputname, ))  # in this case inputname will be an LFN

    return jobNode

################################################################################

  def __generateOutputFiles(self, jobNode, bkLFNs, logFilePath):
    """OutputFile looks like this::

       <OutputFile Name="" TypeName="" TypeVersion="">
         <Parameter Name="" Value=""/>
         ...
         <Replica Location="" Name=""/>
         ....
       </OutputFile>
    """

    if self.eventType is not None:
      eventtype = self.eventType
    else:
      self.log.warn('BookkeepingReport: no eventType specified')
      eventtype = 'Unknown'
    self.log.info('Event type = %s' % (str(self.eventType)))

    outputs = []
    count = 0
    bkTypeDict = {}
    while count < len(self.stepOutputs):
      if 'outputDataName' in self.stepOutputs[count]:
        outputs.append(((self.stepOutputs[count]['outputDataName']),
                        (self.stepOutputs[count]['outputDataType'])))
      if 'outputBKType' in self.stepOutputs[count]:
        bkTypeDict[self.stepOutputs[count]['outputDataName']] = self.stepOutputs[count]['outputBKType']
      count = count + 1
    outputs.append(((self.applicationLog), ('LOG')))
    self.log.info(outputs)
    if isinstance(logFilePath, list):
      logFilePath = logFilePath[0]

    for output, outputtype in list(outputs):
      self.log.info('Looking at output %s %s' % (output, outputtype))
      typeName = outputtype.upper()
      typeVersion = '1'
      fileStats = '0'
      if output in bkTypeDict:
        typeVersion = getOutputType(output, self.stepInputData)[output]
        self.log.info('Setting POOL XML catalog type',
                      'for %s to %s' % (output, typeVersion))
        typeName = bkTypeDict[output].upper()
        self.log.info('Setting explicit BK type version',
                      'for %s to %s and file type to %s' % (output, typeVersion, typeName))

        fileStats, output = self._getFileStatsFromXMLSummary(output, outputtype)

      if not os.path.exists(output):
        self.log.error("Output file does not exist",
                       "Output file name: %s" % output)
        continue
      # Output file size
      if 'size' not in self.step_commons or output not in self.step_commons['size']:
        try:
          outputsize = str(os.path.getsize(output))
        except OSError:
          outputsize = '0'
      else:
        outputsize = self.step_commons['size'][output]

      if 'md5' not in self.step_commons or output not in self.step_commons['md5']:
        comm = 'md5sum ' + str(output)
        resultTuple = systemCall(0, shlex.split(comm))
        status = resultTuple['Value'][0]
        out = resultTuple['Value'][1]
      else:
        status = 0
        out = self.step_commons['md5'][output]

      if status:
        self.log.info("Failed to get md5sum of %s" % str(output))
        self.log.info(str(out))
        md5sum = '000000000000000000000000000000000000'
      else:
        md5sum = out.split()[0]

      if 'guid' not in self.step_commons or output not in self.step_commons['guid']:
        guidResult = getGUID(output)
        guid = ''
        if not guidResult['OK']:
          self.log.error("Could not find GUID",
                         "for %s with message %s" % (output, guidResult['Message']))
        elif guidResult['generated']:
          self.log.warn('PoolXMLFile generated GUID(s) for the following files ',
                        ', '.join(guidResult['generated']))
          guid = guidResult['Value'][output]
        else:
          guid = guidResult['Value'][output]
          self.log.info('Setting POOL XML catalog GUID',
                        'for %s to %s' % (output, guid))
      else:
        guid = self.step_commons['guid'][output]

      if not guid:
        self.log.error('No GUID found', 'for %s' % output)
        raise NameError('No GUID found')

      # find the constructed lfn
      lfn = ''
      if not re.search('.log$', output):
        for outputLFN in bkLFNs:
          if os.path.basename(outputLFN) == output:
            lfn = outputLFN
        if not lfn:
          self.log.error('Could not find LFN', 'for %s' % output)
          raise NameError('Could not find LFN of output file')
      else:
        lfn = '%s/%s' % (logFilePath, self.applicationLog)

      oldTypeName = None
      if 'HIST' in typeName.upper():
        typeVersion = '0'

      # PROTECTION for old production XMLs
      if typeName.upper() == 'HIST':
        typeName = '%sHIST' % (self.applicationName.upper())

      # Add Output to the XML file
      oFileAttributes = (lfn, typeName, typeVersion)
      jobNode, oFile = addChildNode(jobNode, "OutputFile", 1, oFileAttributes)

      # HIST is in the dataTypes e.g. we may have new names in the future ;)
      if oldTypeName:
        typeName = oldTypeName

      if outputtype != 'LOG':
        oFile = addChildNode(oFile, "Parameter", 0, ("EventTypeId", eventtype))
        if fileStats != 'Unknown':
          oFile = addChildNode(oFile, "Parameter", 0, ("EventStat", fileStats))

      oFile = addChildNode(oFile, "Parameter", 0, ("FileSize", outputsize))

      oFile = addChildNode(oFile, "Parameter", 0, ("CreationDate",
                                                   time.strftime('%Y-%m-%d %H:%M:%S',
                                                                 time.gmtime(time.time()))))

      ############################################################
      # Log file replica information
#      if typeName == "LOG":
      if self.applicationLog:
        logfile = self.applicationLog
        if logfile == output:
          logurl = 'http://lhcb-logs.cern.ch/storage'
          url = logurl + logFilePath + '/' + self.applicationLog
          oFile = addChildNode(oFile, "Replica", 0, (url, ))

      oFile = addChildNode(oFile, "Parameter", 0, ("MD5Sum", md5sum))
      oFile = addChildNode(oFile, "Parameter", 0, ("Guid", guid))

    return jobNode

################################################################################

  def _getFileStatsFromXMLSummary(self, output, outputtype):
    """Gets stats per file from the XML summary, considering files registered
    with different cases.

    :params str output: file name looked out
    :params str outputType: file type looked out
    :returns: (str, str) with stats and actual file name
    """
    try:
      return str(self.xf_o.outputsEvents[output]), output
    except AttributeError as e:
      # This happens iff the XML summary can't be created (e.g. for merging MDF files)
      self.log.warn("XML summary not created, unable to determine the output events and setting to 'Unknown'",
                    repr(e))
      return 'Unknown', output
    except KeyError as e:
      self.log.warn("Could not find output LFN in XML summary object",
                    repr(e))

      # here starting to look if by chance the file has been produced with a different case
      for outputFileInXML in self.xf_o.outputsEvents:
        if output.lower() == outputFileInXML.lower():
          self.log.info("Found output LFN in XML summary object with different case",
                        "%s -> %s" % (output, outputFileInXML))
          return str(self.xf_o.outputsEvents[outputFileInXML]), outputFileInXML

      if ('hist' in outputtype.lower()) or ('.root' in outputtype.lower()):
        self.log.warn("HIST file not found in XML summary, event stats set to 'Unknown'",
                      "HIST not found = %s" % output)
        return 'Unknown', output

      raise KeyError("Could not find output LFN in XML summary object")

  def __generateSimulationCondition(self, jobNode):
    """ SimulationCondition looks like this::

         <SimulationCondition>
           <Parameter Name="" Value=""/>
         </SimulationCondition>
    """
    if self.applicationName == "Gauss":
      jobNode, sim = addChildNode(jobNode, "SimulationCondition", 1, ())
      sim = addChildNode(sim, "Parameter", 0, ("SimDescription", self.simDescription))

    return jobNode

################################################################################

  def __getNodeInformation(self):
    """Try to obtain system HostName, CPU, Model, cache and memory.

    This information is not essential to the running of the jobs but
    will be reported if available.
    """
    result = {}
    try:
      result["HostName"] = socket.gethostname()
      result["CPU(MHz)"] = psutil.cpu_freq()[0]
      result["Memory(kB)"] = int(psutil.virtual_memory()[1] / 1024)

      with open("/proc/cpuinfo", "r") as cpuinfo:
        info = cpuinfo.readlines()
      result["ModelName"] = [x.strip().split(":")[1] for x in info if "model name" in x][0].strip()
      result["CacheSize(kB)"] = [x.strip().split(":")[1] for x in info if "cache size" in x][0].strip()

    except BaseException as x:
      self.log.exception("BookkeepingReport failed to obtain node information", lException=x)
      return S_ERROR("Failed to obtain system information")

    return S_OK(result)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Resolve SE takes the workflow SE description and returns the list of
destination storage elements for uploading an output file."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from random import shuffle

from DIRAC import gLogger, gConfig
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

__RCSID__ = "$Id$"


def _setLocalFirst(seList, localSEs):
  """return a shuffled list of SEs from seList, localSEs being first."""
  local = [se for se in seList if se in localSEs]
  remote = [se for se in seList if se not in localSEs]
  shuffle(local)
  shuffle(remote)
  return local + remote


def getDestinationSEList(outputSE, site, outputmode='Any', run=None):
  """Evaluate the output SE list from a workflow and return the concrete list
  of SEs to upload output data."""
  if outputmode.lower() not in ('any', 'local', 'run'):
    raise RuntimeError("Unexpected outputmode")

  if outputmode.lower() == 'run':
    gLogger.verbose("Output mode set to 'run', thus ignoring site parameter")
    if not run:
      raise RuntimeError("Expected runNumber")
    try:
      run = int(run)
    except ValueError as ve:
      raise RuntimeError("Expected runNumber as a number: %s" % ve)

    gLogger.debug("RunNumber = %d" % run)
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    runDestination = TransformationClient().getDestinationForRun(run)
    if not runDestination['OK'] or run not in runDestination['Value']:
      raise RuntimeError("Issue getting destinationForRun (%d): " % run + runDestination.get('Message', 'unknown run'))
    site = runDestination['Value'][run]
    gLogger.verbose("Site set to %s for run %d" % (site, run))
    outputmode = 'Local'

  # Add output SE defined in the job description
  gLogger.info('Resolving workflow output SE description: %s' % outputSE)

  # Check if the SE is defined explicitly for the site
  prefix = site.split('.')[0]
  country = site.split('.')[-1]
  # Concrete SE name
  result = gConfig.getOptions('/Resources/StorageElements/' + outputSE)
  if result['OK']:
    gLogger.info('Found concrete SE %s' % outputSE)
    return [outputSE]

  # Get local SEs
  localSEs = getSEsForSite(site)
  if not localSEs['OK']:
    raise RuntimeError(localSEs['Message'])
  localSEs = localSEs['Value']
  gLogger.verbose("Local SE list is: %s" % (localSEs))

# There is an alias defined for this Site
  associatedSEs = gConfig.getValue('/Resources/Sites/%s/%s/AssociatedSEs/%s' % (prefix, site, outputSE), [])
  if associatedSEs:
    associatedSEs = _setLocalFirst(associatedSEs, localSEs)
    gLogger.info("Found associated SE %s for site %s" % (associatedSEs, site))
    return associatedSEs

  groupSEs = resolveSEGroup(outputSE)
  if not groupSEs:
    raise RuntimeError("Failed to resolve SE " + outputSE)
  gLogger.verbose("Group SE list is: %s" % (groupSEs))

  # Find a local SE or an SE considered as local because the country is associated to it
  if outputmode.lower() == "local":
    # First, check if one SE in the group is local
    for se in localSEs:
      if se in groupSEs:
        gLogger.info("Found eligible local SE: %s" % (se))
        return [se]

    # Final check for country associated SE
    assignedCountry = country
    while True:
      # check if country is already one with associated SEs
      section = '/Resources/Countries/%s/AssociatedSEs/%s' % (assignedCountry, outputSE)
      associatedSEs = gConfig.getValue(section, [])
      if associatedSEs:
        associatedSEs = _setLocalFirst(associatedSEs, localSEs)
        gLogger.info('Found associated SEs %s in %s' % (associatedSEs, section))
        return associatedSEs

      gLogger.verbose("/Resources/Countries/%s/AssignedTo" % assignedCountry)
      opt = gConfig.getOption("/Resources/Countries/%s/AssignedTo" % assignedCountry)
      if opt['OK'] and opt['Value']:
        assignedCountry = opt['Value']
      else:
        # No associated SE and no assigned country, give up
        raise RuntimeError("Could not establish associated SE nor assigned country for country %s" % assignedCountry)

  # For collective Any and All modes return the whole group
  # Make sure that local SEs are passing first
  orderedSEs = _setLocalFirst(groupSEs, localSEs)
  gLogger.info('Found SEs, local first: %s' % orderedSEs)
  return orderedSEs



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""UserStoragePlotter."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.UserStorage import UserStorage

__RCSID__ = "$Id$"


class UserStoragePlotter(BaseReporter):
  """UserStoragePlotter as extension of BaseReporter."""

  _typeName = "UserStorage"
  _typeKeyFields = [dF[0] for dF in UserStorage().definitionKeyFields]

  ##############################################################################
  #
  # Catalog Space
  #

  _reportCatalogSpaceName = "LFN size"

  def _reportCatalogSpace(self, reportRequest):
    """Reports about the LFN size and the catalog space from the accounting."""

    if reportRequest['grouping'] == "StorageElement":
      return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

    _selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (_selectField + ", %s, %s, SUM(%s)/SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'LogicalSize', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "bytes")

    baseDataDict, graphDataDict, __maxValue, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotCatalogSpace(self, reportRequest, plotInfo, filename):
    """Plots about the LFN size and the catalog space."""

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']

    metadata = {'title': "User's LFN space usage by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    plotInfo['graphDataDict'] = self._fillWithZero(span, startTime, endTime, plotInfo['graphDataDict'])
    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  ##############################################################################
  #
  # Catalog Files
  #

  _reportCatalogFilesName = "LFN files"

  def _reportCatalogFiles(self, reportRequest):
    """Reports about the LFN files and the catalog files from the
    accounting."""

    if reportRequest['grouping'] == "StorageElement":
      return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

    _selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (_selectField + ", %s, %s, SUM(%s)/SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'LogicalFiles', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})

    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "files")

    baseDataDict, graphDataDict, __maxValue, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotCatalogFiles(self, reportRequest, plotInfo, filename):
    """Plots about the LFN files and the catalog files."""

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']

    metadata = {'title': "User's Number of LFNs by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    plotInfo['graphDataDict'] = self._fillWithZero(span, startTime, endTime, plotInfo['graphDataDict'])

    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  ##############################################################################
  #
  # Physical Space
  #

  _reportPhysicalSpaceName = "PFN size"

  def _reportPhysicalSpace(self, reportRequest):
    """Reports about the PFN size and the physical space from the
    accounting."""

    _selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (_selectField + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'PhysicalSize', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})

    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "bytes")

    baseDataDict, graphDataDict, __maxValue, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotPhysicalSpace(self, reportRequest, plotInfo, filename):
    """Plots about the PFN size and the physical space."""

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']

    metadata = {'title': "User's PFN space usage by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    plotInfo['graphDataDict'] = self._fillWithZero(span, startTime, endTime, plotInfo['graphDataDict'])

    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  ##############################################################################
  #
  # Physical Files
  #

  _reportPhysicalFilesName = "PFN files"

  def _reportPhysicalFiles(self, reportRequest):
    """Reports about the PFN files and the physical files from the
    accounting."""

    _selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (_selectField + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'PhysicalFiles', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})

    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "files")

    baseDataDict, graphDataDict, __maxValue, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotPhysicalFiles(self, reportRequest, plotInfo, filename):
    """Plots about the PFN files and the physical files."""

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']

    metadata = {'title': "User's Number of PFNs by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    plotInfo['graphDataDict'] = self._fillWithZero(span, startTime, endTime, plotInfo['graphDataDict'])

    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  ##############################################################################
  #
  # PFN vs LFN File Multiplicity
  #

  _reportPFNvsLFNFileMultiplicityName = "PFN/LFN file ratio"

  def _reportPFNvsLFNFileMultiplicity(self, reportRequest):
    """Reports about the PFN/LFN file ratio from the accounting ( only grouped
    by user )."""

    if reportRequest['grouping'] == "User":
      return S_ERROR("Grouping by user when requesting replicas/lfns makes no sense")

    _logicalField = "LogicalFiles"
    _physicalField = "PhysicalFiles"

    return self.__multiplicityReport(reportRequest, _logicalField, _physicalField)

  def _plotPFNvsLFNFileMultiplicity(self, reportRequest, plotInfo, filename):
    """Plots about the PFN/LFN file ratio ( only grouped by user )."""

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']

    metadata = {'title': "User's Ratio of PFN/LFN files by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    plotInfo['graphDataDict'] = self._fillWithZero(span, startTime, endTime, plotInfo['graphDataDict'])

    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  ##############################################################################
  #
  # PFN vs LFN Size Multiplicity
  #

  _reportPFNvsLFNSizeMultiplicityName = "PFN/LFN size ratio"

  def _reportPFNvsLFNSizeMultiplicity(self, reportRequest):
    """Reports about the PFN/LFN size ratio from the accounting ( only grouped
    by user )."""

    if reportRequest['grouping'] == "User":
      return S_ERROR("Grouping by user when requesting replicas/lfns makes no sense")

    _logicalField = "LogicalSize"
    _physicalField = "PhysicalSize"

    return self.__multiplicityReport(reportRequest, _logicalField, _physicalField)

  def _plotPFNvsLFNSizeMultiplicity(self, reportRequest, plotInfo, filename):
    """Plots about the PFN/LFN size ratio ( only grouped by user )."""

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']

    metadata = {'title': "User's Ratio of PFN/LFN space used by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    plotInfo['graphDataDict'] = self._fillWithZero(span, startTime, endTime, plotInfo['graphDataDict'])

    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  ##############################################################################
  #
  # Helpers
  #

  def __multiplicityReport(self, reportRequest, logicalField, physicalField):
    """Calculates the ratio between PFN and LFN."""

    # Step 1 get the total LFNs for each bucket
    selectFields = ("%s, %s, %s, SUM(%s)/SUM(%s)",
                    ['User', 'startTime', 'bucketLength', logicalField, 'entriesInBucket']
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                ('%s', ['User']),
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal
    dataDict, _granularity = retVal['Value']
    self.stripDataField(dataDict, 0)
    bucketTotals = self._getBucketTotals(dataDict)

    # Step 2 get the total PFNs
    _selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (_selectField + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          physicalField, 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)
    # Step 3 divide the PFNs by the total amount of LFNs
    finalData = {}
    for k in dataDict:
      for bt in dataDict[k]:
        if bt in bucketTotals:
          if k not in finalData:
            finalData[k] = {}
          finalData[k][bt] = dataDict[k][bt] / bucketTotals[bt]

    return S_OK({'data': finalData,
                 'graphDataDict': finalData,
                 'granularity': granularity,
                 'unit': 'PFN / LFN'})



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This modules contains utility functions for LHCb DM."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import six
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

__RCSID__ = "$Id$"


def chown(directories, user=None, group=None, mode=None, recursive=False, ndirs=None, fcClient=None):
  """This method may change the user, group or mode of a directory and apply it
  recursively if required."""
  if ndirs is None:
    ndirs = 0
  if not directories:
    return S_OK(ndirs)
  if isinstance(directories, six.string_types):
    directories = directories.split(',')
  if fcClient is None:
    fcClient = FileCatalogClient()
  timeout = 3600 if recursive else 10 * len(directories)
  if user is not None:
    res = fcClient.changePathOwner(dict.fromkeys(directories, user), recursive=recursive, timeout=timeout)
    if not res['OK']:
      res['Action'] = 'changePathOwner'
      return res
  if group is not None:
    res = fcClient.changePathGroup(dict.fromkeys(directories, group), recursive=recursive, timeout=timeout)
    if not res['OK']:
      res['Action'] = 'changePathGroup'
      return res
  if mode is not None:
    res = fcClient.changePathMode(dict.fromkeys(directories, mode), recursive=recursive, timeout=timeout)
    if not res['OK']:
      res['Action'] = 'changePathMode'
      return res

  ndirs += len(directories)
  return S_OK(ndirs)


def createUserDirectory(user):
  """This functions creates (if not existing) a user directory in the DFC."""
  dfc = FileCatalogClient()
  initial = user[0]
  baseDir = os.path.join('/lhcb', 'user', initial, user)
  if dfc.isDirectory(baseDir).get('Value', {}).get('Successful', {}).get(baseDir):
    return S_ERROR('User directory already existing')
  gLogger.info('Creating directory', baseDir)
  res = dfc.createDirectory(baseDir)
  if not res['OK']:
    return res
  gLogger.info('Setting ownership of directory', baseDir)
  return chown(baseDir, user, group='lhcb_user', mode=0o755, recursive=False, fcClient=dfc)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Move files that are Unused or MaxReset from a parent production to its
derived production The argument is a list of productions: comma separated list
of ranges (a range has the form p1:p2)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] [prod1[:prod2][,prod3[:prod4]]' % Script.scriptName, ]))
  Script.registerSwitch('', 'NoReset', "Don't reset the MaxRest files to unused (default is to reset)")
  Script.parseCommandLine(ignoreErrors=True)

  import DIRAC
  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  transClient = TransformationClient()

  resetUnused = True
  switches = Script.getUnprocessedSwitches()
  for switch in switches:
    if switch[0] == 'NoReset':
      resetUnused = False
  args = Script.getPositionalArgs()

  if not len(args):
    print("Specify transformation number...")
    DIRAC.exit(0)
  else:
    ids = args[0].split(",")
    idList = []
    for transId in ids:
      r = transId.split(':')
      if len(r) > 1:
        for i in range(int(r[0]), int(r[1]) + 1):
          idList.append(i)
      else:
        idList.append(int(r[0]))

  for prod in idList:
    res = transClient.getTransformation(prod, extraParams=True)
    if not res['OK']:
      print("Error getting transformation %s" % prod, res['Message'])
    else:
      res = transClient.moveFilesToDerivedTransformation(res['Value'], resetUnused)
      if not res['OK']:
        print("Error updating a derived transformation %d:" % prod, res['Message'])
      else:
        parentProd, movedFiles = res['Value']
        if movedFiles:
          print("Successfully moved files from %d to %d:" % (parentProd, prod))
          for status, val in movedFiles.items():
            print("\t%d files to status %s" % (val, status))


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: DataUsageClient

.. module: DataUsageClient

:synopsis: Class that contains client access to the StorageUsageDB handler.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

# # imports
from DIRAC import S_ERROR
from DIRAC.Core.Base.Client import Client, createClient

__RCSID__ = "$Id$"


@createClient('DataManagement/DataUsage')
class DataUsageClient(Client):
  """
  .. class:: DataUsageClient
  """

  def __init__(self, url=None, **kwargs):
    """c'tor."""
    super(DataUsageClient, self).__init__(**kwargs)
    self.setServer('DataManagement/DataUsage')
    if url:
      self.setServer(url)

  def sendDataUsageReport(self, site, directoryDict, status='New', rpc=None, url='', timeout=120):
    """send data usage report."""
    if not isinstance(directoryDict, dict):
      return S_ERROR('Supplied dictionary is not in correct format!')
    rpcClient = self._getRPC(rpc=rpc, url=url, timeout=timeout)
    return rpcClient.sendDataUsageReport(site, directoryDict, status)

  def getDataUsageSummary(self, startTime, endTime, status, rpc=None, url='', timeout=120):
    """get usage summary."""
    if not (isinstance(startTime, six.string_types) and
            isinstance(endTime, six.string_types) and
            isinstance(status, six.string_types)):
      return S_ERROR('Supplied arguments not in correct format!')
    rpcClient = self._getRPC(rpc=rpc, url=url, timeout=timeout)
    return rpcClient.getDataUsageSummary(startTime, endTime, status)

  def insertToDirMetadata(self, directoryDict, url='', timeout=120):
    """insert metadata to dir or maybe other way around."""
    if not isinstance(directoryDict, dict):
      return S_ERROR('Supplied dictionary is not in correct format!')
    rpcClient = self._getRPC(rpc=None, url=url, timeout=timeout)
    return rpcClient.insertToDirMetadata(directoryDict)

  def getDirMetadata(self, directoryList, url='', timeout=120):
    """get directory metadata."""
    if isinstance(directoryList, six.string_types):
      directoryList = [directoryList]
    elif isinstance(directoryList, (set, tuple, dict)):
      directoryList = list(directoryList)
    elif not isinstance(directoryList, list):
      return S_ERROR('Supplied argument is not in correct format!')
    rpcClient = self._getRPC(rpc=None, url=url, timeout=timeout)
    return rpcClient.getDirMetadata(directoryList)

  def updatePopEntryStatus(self, idList, newStatus, url='', timeout=120):
    """whatever, pop new status."""
    if not isinstance(idList, list) or not isinstance(newStatus, six.string_types):
      return S_ERROR('Supplied arguments are not in correct format!')
    rpcClient = self._getRPC(rpc=None, url=url, timeout=timeout)
    return rpcClient.updatePopEntryStatus(idList, newStatus)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get statistics on number of replicas for a given directory or production."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerBKSwitches()
  dmScript.registerNamespaceSwitches()
  dmScript.registerFileSwitches()

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] [<LFN>] [<LFN>...]' % Script.scriptName, ]))

  Script.registerSwitch("", "Size", "   Get the LFN size [No]")
  Script.registerSwitch('', 'DumpNoReplicas', '   Print list of files without a replica [No]')
  Script.registerSwitch('', 'DumpWithArchives=', '   =<n>, print files with <n> archives')
  Script.registerSwitch('', 'DumpWithReplicas=', '   =<n>, print files with <n> replicas')
  Script.registerSwitch('', 'DumpFailover',
                        '   print files with failover replica (can be used with Dump[With/No]Replicas)')
  Script.registerSwitch('', 'DumpAtSE=', '   print files present at a (list of) SE')
  Script.registerSwitch('', 'DumpNotAtSE=', '   print files absent at a (list of) SE')
  Script.registerSwitch('', 'DumpAtSite=', '   print files present at a (list of) sites')
  Script.registerSwitch('', 'Summary', '   do not print stats per SE nor site')

  Script.parseCommandLine(ignoreErrors=False)

  from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeReplicaStats
  from DIRAC import exit
  exit(executeReplicaStats(dmScript))


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: PopularityAgent

.. module: PopularityAgent

:synopsis: The Popularity Agent creates reports about per LFC directory data usage.

The Popularity Agent creates reports about per LFC directory data usage, based on the
StorageUsageDB/Popularity table. Then it creates an accounting record for each directory,
adding all the relevant directory metadata, obtained from the StorageUsageDB/DirMetadata table.
The accounting records are stored in the AccountingDB and then displayed via the web portal.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# imports
from datetime import datetime, timedelta
# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from LHCbDIRAC.AccountingSystem.Client.Types.Popularity import Popularity
from LHCbDIRAC.DataManagementSystem.Client.DataUsageClient import DataUsageClient
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

__RCSID__ = "$Id$"

AGENT_NAME = "DataManagement/PopularityAgent"


class PopularityAgent(AgentModule):
  """
  .. class:: PopularityAgent
  """
  # # DataUsageClient
  __dataUsageClient = None
  # # StorageUsageDB instance or DMS/DataUsage RPS client
  __stDB = None
  # # BKK Client
  __bkClient = None
  # # work directory
  __workDirectory = None
  # # counter for records to be sent to the accouting
  numPopRows = None

  def initialize(self):
    """agent initialisation."""
    self.am_setOption('PollingTime', 43200)
    if self.am_getOption('DirectDB', False):
      self.__stDB = StorageUsageDB()
      # self.bkClient = BookkeepingClient()#the necessary method is still not available in Bookk. client
    else:
      self.__stDB = DataUsageClient()
    self.__bkClient = BookkeepingClient()
    self.__dataUsageClient = DataUsageClient()
    self.__workDirectory = self.am_getOption("WorkDirectory")
    mkDir(self.__workDirectory)
    self.log.info("Working directory is %s" % self.__workDirectory)
    # by default, collects raw records from Popularity table inserted in the last day
    self.timeInterval = self.am_getOption('timeIntervalForPopularityRecords', 1)
    self.queryTimeout = self.am_getOption('queryTimeout', 3600)
    self.cacheMetadata = {}
    self.limitForCommit = self.am_getOption("LimitForCommit", 1000)

    return S_OK()

# .........................................................................................

  def execute(self):
    """Main loop of Popularity agent."""

    now = datetime.now()
    endTime = datetime(now.year, now.month, now.day, 0, 0, 0)
    startTime = endTime - timedelta(days=self.timeInterval)
    endTimeQuery = endTime.isoformat()
    startTimeQuery = startTime.isoformat()
    # query all traces in popularity in the time rage startTime,endTime and status =new
    # the condition to get th etraces is the AND of the time range and the status new
    self.log.info("Querying Pop db to retrieve entries in time range %s - %s " %
                  (startTimeQuery, endTimeQuery))
    status = 'New'
    res = self.__dataUsageClient.getDataUsageSummary(startTimeQuery, endTimeQuery, status,
                                                     timeout=self.queryTimeout)
    if not res['OK']:
      self.log.error("Error querying Popularity table.. %s" % res['Message'])
      return S_ERROR(res['Message'])
    val = res['Value']
    self.log.info("Retrieved %d entries from Popularity table" % len(val))
    # Build popularity report, and store the Ids in a  list:
    idList = set()
    traceDict = {}
    for row in val:
      self.log.debug("row: %s" % str(row))
      rowId, dirLfn, site, count, insertTime = row
      if rowId not in idList:
        idList.add(rowId)
      else:
        self.log.error("Same Id found twice! %d " % rowId)
        continue
      if dirLfn.startswith('/lhcb/user/'):
        self.log.verbose("Private user directory. No metadata stored in Bkk %s " % dirLfn)
        continue
      # get the day (to do )
      dayBin = (insertTime - startTime).days
      traceDict[dayBin][dirLfn][site] = \
          traceDict.setdefault(dayBin, {}).setdefault(dirLfn, {}).setdefault(site, 0) + count

    # print a summary
    dayList = sorted(traceDict)
    for day in dayList:
      self.log.info(" ###### day %s (starting from %s ) " % (day, startTimeQuery))
      self.log.info("---- %d directories touched:" % len(traceDict[day]))
      for lfn in traceDict[day]:
        self.log.verbose(" ---- lfn %s " % lfn)
        for site in traceDict[day][lfn]:
          self.log.verbose(" -------- site  %s  count: %d " % (site, traceDict[day][lfn][site]))

    self.log.info("Retrieve meta-data information for each directory ")
    now = Time.dateTime()
    self.numPopRows = 0  # keep a counter of the records to send to accounting data-store
    for day in traceDict:
      timeForAccounting = self.computeTimeForAccounting(startTime, day)
      self.log.info("Processing day %s - time for accounting %s " % (day, timeForAccounting))
      for dirLfn in traceDict[day]:
        # did = configName = configVersion = conditions = processingPass = eventType = fileType = production = "na"
        # retrieve the directory meta-data from the DirMetadata table
        self.log.info("Processing dir %s " % dirLfn)

        metaForDir = self.cacheMetadata.get(dirLfn)
        if not metaForDir:
          dirList = [dirLfn]
          # this could be done in a bulk query for a list of directories... TBD
          res = self.__dataUsageClient.getDirMetadata(dirList)
          if not res['OK']:
            self.log.error("Error retrieving directory meta-data %s " % res['Message'])
            continue
          dirMetadata = res['Value'].get(dirLfn)
          if not res['Value'] or not dirMetadata:
            self.log.info("Cache missed: query BK to retrieve '%s' metadata and store  cache" % dirList)
            res = self.__bkClient.getDirectoryMetadata(dirList)
            if not res['OK']:
              self.log.error("Failed to query Bookkeeping %s" % res['Message'])
              metadata = None
            else:
              self.log.verbose("Successfully queried Bookkeeping, result: %s " % res)
              metadata = res['Value'].get('Successful', {}).get(dirLfn, [{}])[0]
            if not metadata:
              self.log.warn("Directory is not registered in Bookkeeping! %s " % dirLfn)
              configName = configVersion = conditions = processingPass = eventType = fileType = production = "na"
            else:
              metadata['Visibility'] = metadata.pop('VisibilityFlag', metadata.get('Visibility', 'na'))
              configName = metadata['ConfigName']
              configVersion = metadata['ConfigVersion']
              conditions = metadata['ConditionDescription']
              processingPass = metadata['ProcessingPass']
              eventType = metadata['EventType']
              fileType = metadata['FileType']
              production = metadata['Production']

              self.log.info("Cache this entry in DirMetadata table..")
              res = self.__dataUsageClient.insertToDirMetadata({dirLfn: metadata})
              if not res['OK']:
                self.log.error("Failed to insert metadata in DirMetadata table! %s " % res['Message'])
              else:
                self.log.info("Successfully inserted metadata for directory %s in DirMetadata table " % dirLfn)
                self.log.verbose("result: %s " % res)

          else:
            self.log.info("Directory %s was cached in DirMetadata table" % dirLfn)
            try:
              __did, configName, configVersion, conditions, \
                  processingPass, eventType, fileType, production = dirMetadata[0:8]
            except BaseException:
              self.log.error("Error decoding directory cached information", dirMetadata)
              continue
          self.cacheMetadata[dirLfn] = (
              configName,
              configVersion,
              conditions,
              processingPass,
              eventType,
              fileType,
              production)
        else:
          configName, configVersion, conditions, processingPass, eventType, fileType, production = metaForDir

        for site in traceDict[day][dirLfn]:
          usage = traceDict[day][dirLfn][site]
          # compute the normalized usage, dividing by the number of files in the directory:
          normUsage = usage  # to be done! after we have decided how to normalize
          # Build record for the accounting
          popRecord = Popularity()
          popRecord.setStartTime(timeForAccounting)
          popRecord.setEndTime(timeForAccounting)
          popRecord.setValueByKey("DataType", configName)
          popRecord.setValueByKey("Activity", configVersion)
          popRecord.setValueByKey("FileType", fileType)
          popRecord.setValueByKey("Production", production)
          popRecord.setValueByKey("ProcessingPass", processingPass)
          popRecord.setValueByKey("Conditions", conditions)
          popRecord.setValueByKey("EventType", eventType)
          popRecord.setValueByKey("StorageElement", site)
          popRecord.setValueByKey("Usage", usage)
          popRecord.setValueByKey("NormalizedUsage", normUsage)
          res = gDataStoreClient.addRegister(popRecord)
          if not res['OK']:
            self.log.error("ERROR: addRegister returned: %s" % res['Message'])
            continue
          self.numPopRows += 1
          self.log.info(">>> Sending record to accounting for: %s %s %s %s %s %s %s %s %s %d %d " %
                        (timeForAccounting, configName, configVersion, fileType,
                         production, processingPass, conditions, eventType, site, usage, normUsage))
          if self.numPopRows > self.limitForCommit:
            res = self.__commitAccounting()
            if not res['OK']:
              return res
    # then set the status to Used
    res = self.__commitAccounting()
    if not res['OK']:
      return res
    self.log.info("Set the status to Used for %d entries" % len(idList))
    from DIRAC.Core.Utilities.List import breakListIntoChunks
    for idChunk in breakListIntoChunks(list(idList), 1000):
      res = self.__dataUsageClient.updatePopEntryStatus(list(idChunk), 'Used', timeout=self.queryTimeout)
      if not res['OK']:
        self.log.error("Error to update status in  Popularity table.. %s" % res['Message'])
        return res
    self.log.info("Status updated to Used correctly for %s entries " % len(idList))

    return S_OK()

# .........................................................................................

  def __commitAccounting(self):
    res = gDataStoreClient.commit()
    if not res['OK']:
      self.log.error("while committing %d Popularity records" % self.numPopRows, res['Message'])
    else:
      self.log.info("%s records for Popularity type successfully committed" % self.numPopRows)
      self.numPopRows = 0
    return res

  def computeTimeForAccounting(self, startTime, day):
    """Compute the time for the accounting record, starting from the start time
    of the query and the day bin."""
    self.log.verbose("find time for accounting for startTime: %s + day %s " % (startTime, day))
    daysToAdd = timedelta(days=day, hours=12)  # add 12h just to put the value in the middle of time bin
    self.log.verbose("timedelta to add: %s " % daysToAdd)
    accTime = startTime + daysToAdd
    self.log.verbose("accTime = %s " % accTime)
    return accTime



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve files of a BK query."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerBKSwitches()
  Script.registerSwitch('', 'File=', '   Provide a list of BK paths')
  Script.registerSwitch('', 'Term', '   Provide the list of BK paths from terminal')
  Script.registerSwitch('', 'Output=', '  Specify a file that will contain the list of files')
  Script.registerSwitch('', 'OptionsFile=', '   Create a Gaudi options file')
  maxFiles = 20
  Script.registerSwitch(
      '',
      'MaxFiles=', '   Print only <MaxFiles> lines on stdout (%d if output, else All)' % maxFiles)
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... ' % Script.scriptName]))

  Script.parseCommandLine()

  from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeGetFiles
  executeGetFiles(dmScript, maxFiles)


if __name__ == "__main__":
  main()
